// ************************************************************************* //
/// AUTHOR: Alexis Angelidis.
/// STARTED: Fri Nov  7 19:13:36 NZDT 2003.
/// DESCRIPTION:
/// This file contains all the very basic 3d types:
///  Point2d, Vector2d, Point3i, Vector3i, Point3d, Vector3d, BBox3d,
///  Quaternion, Matrix2i, Matrix3d, Matrix4d, Translation, Rotation, Scale,
///   ScaleT.
// ************************************************************************* //
#ifndef	__SWPTYPES_HH__
#define	__SWPTYPES_HH__

// ************************************************************************* //
#ifndef WIN32
extern "C"
{
#endif
#include <math.h>
#include <stdlib.h>
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#ifndef WIN32
}
#endif

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795
#endif

// ************************************************************************* //
#include <iostream>

// ************************************************************************* //
#ifndef RAD2DEG
#define RAD2DEG(a) ((a) * 57.2957795130823208768)
#endif//RAD2DEG
#ifndef DEG2RAD
#define DEG2RAD(a) ((a) * 0.01745329251994329576)
#endif//DEG2RAD

// ************************************************************************* //
namespace
{

// ************************************************************************* //
inline void	nrerror(const char *msg)
{
  std::cout << msg << std::endl;
  std::exit(1);
}

// ************************************************************************* //
inline double	myacos(const double costheta)
{
  return (1.0 <= costheta)? 0.0: ((costheta <= -1.0)? M_PI: acos(costheta));
}

// ************************************************************************* //
template <class T> inline
T	sq(const T t)
{
  return t * t;
}

// ************************************************************************* //
/*
template <class T> inline
const T	max(const T a,
	    const T b)
{
  return a < b ? b: a;
}
*/

// ************************************************************************* //
/*
template <class T> inline
T	min(const T a,
	    const T b)
{
  return a < b ? a: b;
}
*/

// ************************************************************************* //
template <class T> inline
T	bound(const T x,
	      const T a,
	      const T b)
{
  return (x <= a) ? a: ((b <= x)? b: x);
}

// ************************************************************************* //
}

// ************************************************************************* //
/// DESCRIPTION: Classes prototypes.
// ************************************************************************* //
class Vector2d;
class Point2d;
class Vector3i;
class Point3i;
class Vector3d;
class Point3d;
class BBox3d;
class Quaternion;
class Matrix2i;
class Matrix3d;
class Matrix4d;
class Transformation;
class Translation;
class Rotation;
class UnitRotation;
class Scale;
class ScaleT;

// ************************************************************************* //
/// DESCRIPTION: Basic Vector2d class.
// ************************************************************************* //
struct Vector2d
{
  double	x, y;
public:
  Vector2d() {}
  Vector2d(const double x_, const double y_): x(x_), y(y_) {}
  explicit Vector2d(const Point2d &p);
  Vector2d	operator*(const double d) const
  { return Vector2d(d * x, d * y); }
  double	operator*(const Vector2d &v) const 
  { return x * v.x + y * v.y; }
  Vector2d	&operator*=(const double d) { x *= d; y *= d; return *this; }
  double	norm2() const { return x * x + y * y; }
  double	norm() const { return sqrt(norm2()); }
  double	operator()(const int i) const { return i? y: x; }
  double	&operator()(const int i) { return i? y: x; }
  void		eqUnit() { operator*=(1.0 / norm()); }
  Vector2d	unit() const { return operator*(1.0 / norm()); }
};

// ************************************************************************* //
/// DESCRIPTION: Basic Point2d class.
// ************************************************************************* //
struct Point2d
{
  double	x, y;
public:
  Point2d() {}
  Point2d(const double d): x(d), y(d) {}
  Point2d(const double x_, const double y_): x(x_), y(y_) {}
  explicit Point2d(const Point3d &p);
  void		glVertex() const { glVertex2d(x, y); }
  Vector2d	operator-(const Point2d &p) const
  { return Vector2d(x - p.x, y - p.y); }
  Point2d	operator+(const Vector2d &v) const
  { return Point2d(x + v.x, y + v.y); }
  Point2d	operator-(const Vector2d &v) const
  { return Point2d(x - v.x, y - v.y); }
  Point2d	&operator-=(const Vector2d &v)
  { x -= v.x; y -= v.y; return *this; }
  Point2d	&operator+=(const Vector2d &v)
  { x += v.x; y += v.y; return *this; }
  double	operator()(const int i) const { return i? y: x; }
  double	&operator()(const int i) { return i? y: x; }
  friend std::istream	&operator>>(std::istream &is, Point2d &p)
  { return is >> p.x >> p.y; }
  friend std::ostream	&operator<<(std::ostream &os, const Point2d &p)
  { return os << p.x << " " << p.y; }
  double	norm() const { return sqrt(x * x + y * y); }
  double	norm2() const { return x * x + y * y; }
  Point2d	proj2line(const Point2d &a, const Point2d &b) const;
};

// ************************************************************************* //
inline Vector2d::Vector2d(const Point2d &p): x(p.x), y(p.y)
{
}

// ************************************************************************* //
/// DESCRIPTION: Projects point on line (a, b).
// ************************************************************************* //
inline Point2d	Point2d::proj2line(const Point2d	&a,
				   const Point2d	&b) const
{
  const Vector2d	m(b - a);

  return a + m * (((*this - a) * m) / (a - b).norm2());
}

// ************************************************************************* //
/// DESCRIPTION: Basic Vector3d class.
// ************************************************************************* //
struct Vector3d
{
  double	x, y, z;
public:
  Vector3d() {}
  Vector3d(const double d): x(d), y(d), z(d) {}
  Vector3d(const double x_, const double y_, const double z_):
    x(x_), y(y_), z(z_) {}
  explicit Vector3d(const Point3d &p);
  Vector3d	&operator+=(const Vector3d &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Vector3d	&operator-=(const Vector3d &v)
  { x -= v.x; y -= v.y; z -= v.z; return *this; }
  Vector3d	&operator*=(const double d)
  { x *= d; y *= d; z *= d; return *this; }
  Vector3d	&operator/=(const double d)
  { x /= d; y /= d; z /= d; return *this; }
  friend Vector3d	operator*(const double d, const Vector3d &v)
  { return Vector3d(d * v.x, d * v.y, d * v.z); }
  Vector3d	operator+(const Vector3d &v) const
  { return Vector3d(x + v.x, y + v.y, z + v.z); }
  Vector3d	operator-(const Vector3d &v) const
  { return Vector3d(x - v.x, y - v.y, z - v.z); }
  double	operator*(const Vector3d &v) const
  { return x * v.x + y * v.y + z * v.z; }
  Vector3d	operator/(const double d) const
  { return Vector3d(x / d, y / d, z / d); }
  Vector3d	operator*(const double d) const
  { return Vector3d(x * d, y * d, z * d); }
  double &operator()(const unsigned i) { return i? ((i == 2)? z: y): x; }
  double operator()(const unsigned i) const { return i? ((i == 2)? z: y): x; }
  double &operator[](const unsigned i) { return i? ((i == 2)? z: y): x; }
  double operator[](const unsigned i) const { return i? ((i == 2)? z: y): x; }
  friend std::istream	&operator>> (std::istream &is, Vector3d &v);
  friend std::ostream	&operator<<(std::ostream &os, const Vector3d &v);
  Vector3d	operator-() const { return Vector3d(-x, -y, -z); }
  Vector3d	&operator*=(const Matrix4d &m);
  Vector3d	&operator*=(const Matrix3d &m);
  double	norm() const { return sqrt(x * x + y * y + z * z); }
  double	norm2() const { return x * x + y * y + z * z; }
  void		eqUnit() { operator*=(1.0 / norm()); }
  Vector3d	unit() const { return operator*(1.0 / norm()); }
  Vector3d	operator^(const Vector3d &v) const
  { return Vector3d(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x); }
  void		glVertex() const { glVertex3d(x, y, z); }
  void		glNormal() const { glNormal3d(x, y, z); }
  void		glColor() const { glColor3d(x, y, z); }
  void		eqSymx() { x = -x; }
  void		eqSymy() { y = -y; }
  void		eqSymz() { z = -z; }
  void		eqSym(const unsigned sym);
  Vector3d	symx() const { return Vector3d(-x, y, z); }
  Vector3d	symy() const { return Vector3d(x, -y, z); }
  Vector3d  	symz() const { return Vector3d(x, y, -z); }
  Vector3d	sym(const unsigned sym) const;
};

// ************************************************************************* //
/// DESCRIPTION: assign to the point one of its symmetries.
// ************************************************************************* //
inline void	Vector3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ************************************************************************* //
/// DESCRIPTION: Returns symmetries of the point.
// ************************************************************************* //
inline Vector3d	Vector3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    Vector3d		&v)
{
  is.read(reinterpret_cast<char *>(&v.x), 8);
  is.read(reinterpret_cast<char *>(&v.y), 8);
  is.read(reinterpret_cast<char *>(&v.z), 8);
  return is;
}

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Vector3d	&v)
{
  os.write(reinterpret_cast<const char *>(&v.x), 8);
  os.write(reinterpret_cast<const char *>(&v.y), 8);
  os.write(reinterpret_cast<const char *>(&v.z), 8);
  return os;
}

// ************************************************************************* //
/// DESCRIPTION: Basic Point3d class.
// ************************************************************************* //
struct Point3d
{
  double	x, y, z;
public:
  Point3d() {}
  Point3d(const double d): x(d), y(d), z(d) {}
  Point3d(const double x_, const double y_, const double z_):
    x(x_), y(y_), z(z_) {}
  explicit Point3d(const Vector3d &v): x(v.x), y(v.y), z(v.z) {}
  void		glVertex() const { glVertex3d(x, y, z); }
  void		glColor() const { glColor3d(x, y, z); }
  double	&operator()(const unsigned i)
  { return i? ((i == 2)? z: y): x; }
  double	operator()(const unsigned i) const
  { return i? ((i == 2)? z: y): x; }
  double	&operator[](const unsigned i)
  { return i? ((i == 2)? z: y): x; }
  double	operator[](const unsigned i) const
  { return i? ((i == 2)? z: y): x; }
  Vector3d	operator-(const Point3d &p) const
  { return Vector3d(x - p.x, y - p.y, z - p.z); }
  Point3d	operator+(const Point3d &p) const
  { return Point3d(x + p.x, y + p.y, z + p.z); }
  Point3d	operator+(const Vector3d &v) const
  { return Point3d(x + v.x, y + v.y, z + v.z); }
  Point3d	operator-(const Vector3d &v) const
  { return Point3d(x - v.x, y - v.y, z - v.z); }
  Point3d	operator*(const double d) const
  { return Point3d(x * d, y * d, z * d); }
  Point3d	&operator-=(const Vector3d &v)
  { x -= v.x; y -= v.y; z -= v.z; return *this; }
  Point3d	&operator+=(const Vector3d &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Point3d	&operator+=(const Point3d &p)
  { x += p.x; y += p.y; z += p.z; return *this; }
  Point3d	&operator*=(const double d)
  { x *= d; y *= d; z *= d; return *this; }
  Point3d	&operator/=(const double d)
  { x /= d; y /= d; z /= d; return *this; }
  Point3d	&operator*=(const Matrix4d &m);
  Point3d	operator-() const { return Point3d(-x, -y, -z); }
  friend std::istream	&operator>>(std::istream &is, Point3d &p);
  friend std::ostream	&operator<<(std::ostream &os, const Point3d &p);
  void		eqJitter(const double max);
  void		glTranslate() const { glTranslated(x, y, z); }
  void		eqSphereIntersection(const Point3d &s, const Vector3d &d,
				     const Point3d &c, const double r);
  Point3d	proj2planeAlong(const Point3d &o, const Vector3d &n,
				const Vector3d &dir) const;

  void		eqSymx() { x = -x; }
  void		eqSymy() { y = -y; }
  void		eqSymz() { z = -z; }
  void		eqSym(const unsigned sym);
  Point3d	symx() const { return Point3d(-x, y, z); }
  Point3d	symy() const { return Point3d(x, -y, z); }
  Point3d  	symz() const { return Point3d(x, y, -z); }
  Point3d	sym(const unsigned sym) const;
};

// ************************************************************************* //
/// DESCRIPTION: assign to the point one of its symmetries.
// ************************************************************************* //
inline void	Point3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ************************************************************************* //
/// DESCRIPTION: Returns symmetries of the point.
// ************************************************************************* //
inline Point3d	Point3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ************************************************************************* //
inline std::istream	&operator>>(std::istream	&is,
				    Point3d		&p)
{ 
  is.read(reinterpret_cast<char *>(&p.x), 8);
  is.read(reinterpret_cast<char *>(&p.y), 8);
  is.read(reinterpret_cast<char *>(&p.z), 8);
  return is;
}

// ************************************************************************* //
inline std::ostream	&operator<<(std::ostream	&os,
				    const Point3d	&p)
{
  os.write(reinterpret_cast<const char *>(&p.x), 8);
  os.write(reinterpret_cast<const char *>(&p.y), 8);
  os.write(reinterpret_cast<const char *>(&p.z), 8);
  return os;
}

// ************************************************************************* //
/// DESCRIPTION: Projects point on plane defined by point o and normal n,
///   along vector dir.
// ************************************************************************* //
inline Point3d	Point3d::proj2planeAlong(const Point3d	&o,
					 const Vector3d	&n,
					 const Vector3d	&dir) const
{
  return *this - dir * (((*this - o) * n) / (dir * n));
}

// ************************************************************************* //
inline Point2d::Point2d(const Point3d &p): x(p.x), y(p.y)
{
}

// ************************************************************************* //
inline void	Point3d::eqSphereIntersection(const Point3d	&s,
					      const Vector3d	&d,
					      const Point3d	&center,
					      const double	r)
{
  Vector3d	s2(s - center);
  const double	a(sq(d(0)) + sq(d(1)) + sq(d(2)));
  const double	b(2.0 * (s2(0) * d(0) + s2(1) * d(1) + s2(2) * d(2)));
  const double	c(sq(s2(0)) + sq(s2(1)) + sq(s2(2)) - sq(r));
  const double	delta(sq(b) - 4.0 * a * c);
  const double	t1((-b + sqrt(delta)) / (2.0 * a));
  const double	t2((-b - sqrt(delta)) / (2.0 * a));
  const double	t(min(t1, t2));

  *this = s + t * d;
}

// ************************************************************************* //
inline void	Point3d::eqJitter(const double max)
{
  const double	scale(max / RAND_MAX);
  const double	trans(max * 0.5);

  x += (scale * rand()) - trans;
  y += (scale * rand()) - trans;
  z += (scale * rand()) - trans;
}

// ************************************************************************* //
inline Vector3d::Vector3d(const Point3d &p): x(p.x), y(p.y), z(p.z)
{
}

// ************************************************************************* //
/// DESCRIPTION: Basic Vector3i class.
// ************************************************************************* //
struct Vector3i
{
  int	x, y, z;
public:
  Vector3i() {}
  Vector3i(const int d): x(d), y(d), z(d) {}
  Vector3i(const int x_, const int y_, const int z_):
    x(x_), y(y_), z(z_) {}
  explicit Vector3i(const Point3i &p);
  Vector3i	&operator+=(const Vector3i &v)
  { x += v.x; y += v.y; z += v.z; return *this; }
  Vector3i	&operator*=(const int d)
  { x *= d; y *= d; z *= d; return *this; }
  Vector3i	&operator/=(const int d)
  { x /= d; y /= d; z /= d; return *this; }
  friend Vector3i	operator*(const int d, const Vector3i &v)
  { return Vector3i(d * v.x, d * v.y, d * v.z); }
  Vector3i	operator+(const Vector3i &v) const
  { return Vector3i(x + v.x, y + v.y, z + v.z); }
  Vector3i	operator-(const Vector3i &v) const
  { return Vector3i(x - v.x, y - v.y, z - v.z); }
  int	operator*(const Vector3i &v) const
  { return x * v.x + y * v.y + z * v.z; }
  Vector3i	operator/(const int d) const
  { return Vector3i(x / d, y / d, z / d); }
  Vector3i	operator*(const int d) const
  { return Vector3i(x * d, y * d, z * d); }
  int &operator()(const unsigned i) { return i? ((i == 2)? z: y): x; }
  int operator()(const unsigned i) const { return i? ((i == 2)? z: y): x; }
  int &operator[](const unsigned i) { return i? ((i == 2)? z: y): x; }
  int operator[](const unsigned i) const { return i? ((i == 2)? z: y): x; }
  friend std::ostream	&operator<<(std::ostream &os, const Vector3i &v)
  { return os << v.x << " " << v.y << " " << v.z; }
  Vector3i	operator-() const { return Vector3i(-x, -y, -z); }
  Vector3i	operator^(const Vector3i &v) const
  { return Vector3i(y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x); }
  void		glVertex() const { glVertex3i(x, y, z); }
  void		glNormal() const { glNormal3i(x, y, z); }
};

// ************************************************************************* //
/// DESCRIPTION: Basic Point3i class.
// ************************************************************************* //
struct Point3i
{
  int	x, y, z;
public:
  Point3i() {}
  Point3i(const int d): x(d), y(d), z(d) {}
  Point3i(const int x_, const int y_, const int z_):
    x(x_), y(y_), z(z_) {}
  void		glVertex() const { glVertex3i(x, y, z); }
  Vector3i	operator-(const Point3i &p) const
  { return Vector3i(x - p.x, y - p.y, z - p.z); }
  friend std::istream	&operator>>(std::istream &is, Point3i &p)
  { return is >> p.x >> p.y >> p.z; }
  friend std::ostream	&operator<<(std::ostream &os, const Point3i &p)
  { return os << p.x << " " << p.y << " " << p.z; }
};

// ************************************************************************* //
inline Vector3i::Vector3i(const Point3i &p): x(p.x), y(p.y), z(p.z)
{
}

// ************************************************************************* //
struct BBox3d
{
  Point3d	pmin, pmax;
public:
  BBox3d() {}
  BBox3d(const Point3d &qmin, const Point3d &qmax):
    pmin(qmin), pmax(qmax) {}
  BBox3d(const Point3d &q): pmin(q), pmax(q) {}
  BBox3d	&operator*=(const Matrix4d &m);
  void		fill(const Point3d &p);
  void		fill(const BBox3d &b);
  void		expand(const double d);
  void		glRender() const;
  Point3d	corner(const char i) const;
  bool		contains(const Point3d &p) const;
  friend std::ostream	&operator<<(std::ostream &os, const BBox3d &b)
  { return os << b.pmin << " " << b.pmax; }
  void		eqSymx();
  void		eqSymy();
  void		eqSymz();
  void		eqSym(const unsigned sym);
  BBox3d	symx() const;
  BBox3d	symy() const;
  BBox3d  	symz() const;
  BBox3d	sym(const unsigned sym) const;
  bool		intersects(const BBox3d &b) const;
};

// ************************************************************************* //
inline void	BBox3d::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ************************************************************************* //
inline BBox3d	BBox3d::sym(const unsigned sym) const
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return *this;
}

// ************************************************************************* //
inline bool	BBox3d::intersects(const BBox3d &b) const
{
  return b.pmin(0) <= pmax(0) && pmin(0) <= b.pmax(0) && 
    b.pmin(1) <= pmax(1) && pmin(1) <= b.pmax(1) &&
    b.pmin(2) <= pmax(2) && pmin(2) <= b.pmax(2);
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ************************************************************************* //
inline void	BBox3d::eqSymx()
{
  const double	tmp(pmin(0));

  pmin(0) = -pmax(0);
  pmax(0) = -tmp;
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relatively to plane y = 0.
// ************************************************************************* //
inline void	BBox3d::eqSymy()
{
  const double	tmp(pmin(1));

  pmin(1) = -pmax(1);
  pmax(1) = -tmp;
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relatively to plane z = 0.
// ************************************************************************* //
inline void	BBox3d::eqSymz()
{
  const double	tmp(pmin(2));

  pmin(2) = -pmax(2);
  pmax(2) = -tmp;
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ************************************************************************* //
inline BBox3d	BBox3d::symx() const
{
  return BBox3d(Point3d(-pmax(0), pmin(1), pmin(2)),
		Point3d(-pmin(0), pmax(1), pmax(2)));
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ************************************************************************* //
inline BBox3d	BBox3d::symy() const
{
  return BBox3d(Point3d(pmin(0), -pmax(1), pmin(2)),
		Point3d(pmax(0), -pmin(1), pmax(2)));
}

// ************************************************************************* //
/// DESCRIPTION: Turns this into a bbox symetric relative to plane x = 0.
// ************************************************************************* //
inline BBox3d	BBox3d::symz() const
{
  return BBox3d(Point3d(pmin(0), pmin(1), -pmax(2)),
		Point3d(pmax(0), pmax(1), -pmin(2)));
}

/* ************************************************************************* */
inline void	BBox3d::fill(const Point3d &p)
{
  pmin.x = min(p.x, pmin.x);
  pmin.y = min(p.y, pmin.y);
  pmin.z = min(p.z, pmin.z);
  pmax.x = max(p.x, pmax.x);
  pmax.y = max(p.y, pmax.y);
  pmax.z = max(p.z, pmax.z);
}

// ************************************************************************* //
inline void	BBox3d::fill(const BBox3d &b)
{
  fill(b.pmin);
  fill(b.pmax);
}

// ************************************************************************* //
inline void	BBox3d::expand(const double d)
{
  pmin.x -= d;
  pmin.y -= d;
  pmin.z -= d;
  pmax.x += d;
  pmax.y += d;
  pmax.z += d;
}

// ************************************************************************* //
inline Point3d	BBox3d::corner(const char i) const
{
  return Point3d((i & 0x1)? pmin.x: pmax.x,
		 (i & 0x2)? pmin.y: pmax.y,
		 (i & 0x4)? pmin.z: pmax.z);
}

// ************************************************************************* //
inline bool	BBox3d::contains(const Point3d &p) const
{
  return pmin(0) <= p(0) && pmin(1) <= p(1) && pmin(2) <= p(2) &&
    p(0) < pmax(0) && p(1) < pmax(1) && p(2) < pmax(2);
}

// ************************************************************************* //
inline void	BBox3d::glRender() const
{
  glBegin(GL_LINE_STRIP);
  pmax.glVertex();
  Point3d(pmax.x, pmin.y, pmax.z).glVertex();
  Point3d(pmin.x, pmin.y, pmax.z).glVertex();
  pmin.glVertex();
  Point3d(pmax.x, pmin.y, pmin.z).glVertex();
  Point3d(pmax.x, pmax.y, pmin.z).glVertex();
  Point3d(pmin.x, pmax.y, pmin.z).glVertex();
  Point3d(pmin.x, pmax.y, pmax.z).glVertex();
  pmax.glVertex();
  Point3d(pmax.x, pmax.y, pmin.z).glVertex();
  glEnd();//GL_LINE_STRIP
  glBegin(GL_LINES);
  Point3d(pmax.x, pmin.y, pmax.z).glVertex();
  Point3d(pmax.x, pmin.y, pmin.z).glVertex();
  Point3d(pmin.x, pmax.y, pmax.z).glVertex();
  Point3d(pmin.x, pmin.y, pmax.z).glVertex();
  Point3d(pmin.x, pmax.y, pmin.z).glVertex();
  pmin.glVertex();
  glEnd();//GL_LINES
}

// ************************************************************************* //
/// DESCRIPTION: Virtual transformation class.
///  Can transform points and normals (when dealing with surfaces).
///  Transformation can be alibi (relative, i.e. cartesian tensor) or alias 
///  (absolute).
// ************************************************************************* //
class Transformation
{
public:
  virtual Transformation	*copy() const = 0;
  virtual Matrix4d		asMatrix() const = 0;
  virtual Matrix4d		operator*(const Matrix4d &m) const;
  virtual Matrix4d		operator*(const Transformation &t) const;
  virtual Transformation	&operator^=(const double t) = 0;
  virtual void			operator()(Point3d &p, Vector3d &v) const = 0;
  virtual Point3d		&operator()(Point3d &p) const = 0;
  virtual Point3d		&operator()(Point3d &p, 
					    const double t) const = 0;
  virtual void			cast(const Transformation &t) = 0;
  virtual unsigned		nbSteps(const BBox3d &b,
					const double k) const = 0;
  virtual void		eqSymx() = 0;
  virtual void		eqSymy() = 0;
  virtual void		eqSymz() = 0;
  virtual void		eqSym(const unsigned sym);
  virtual void		eqInverse() = 0;
  virtual double	det() const = 0;
  virtual Matrix4d	mlog() const = 0;
};

// ************************************************************************* //
inline void	Transformation::eqSym(const unsigned sym)
{
  if (sym == 1)
    eqSymx();
  else if (sym == 2)
    eqSymy();
  else if (sym == 3)
    eqSymz();
}

// ************************************************************************* //
/// DESCRIPTION: Rotation around the origin.
// ************************************************************************* //
class Quaternion: public Transformation
{
  friend class	Rotation;
  double	_m_s;
  Vector3d	_m_v;
public:
  Quaternion() {}
  Quaternion(const double d): _m_s(d), _m_v(d) {}
  Quaternion(const double s, const Vector3d &v): _m_s(s), _m_v(v) {}
  Quaternion(const double s, const double x, const double y, const double z):
    _m_s(s), _m_v(x, y, z) {}
  Quaternion(const Vector3d &v): _m_s(0.0), _m_v(v) {}
  Transformation	*copy() const { return new Quaternion(*this); }
  Quaternion	&operator+=(const Quaternion &q)
  { _m_s += q._m_s; _m_v += q._m_v; return *this; }
  Quaternion	&operator*=(const double d) 
  { _m_s *= d; _m_v *= d; return *this; }
  Quaternion	&operator/=(const double d)
  { _m_s /= d; _m_v /= d; return *this; }
  Quaternion	operator-() const { return Quaternion(-_m_s, -_m_v); }
  Quaternion	operator^(const double d) const;
  Quaternion	&operator^=(const double d);
  Quaternion	operator*(const double d) const 
  { return Quaternion(_m_s * d, _m_v * d); }
  Matrix4d	asMatrix() const;
  friend Quaternion	operator*(const double d, const Quaternion &q)
  { return q * d; }
  Quaternion	operator/(const double d) const 
  { return Quaternion(_m_s / d, _m_v / d); }
  Quaternion	operator*(const Quaternion &q) const 
  { return Quaternion(_m_s * q._m_s - _m_v * q._m_v,
		      _m_s * q._m_v + q._m_s * _m_v + (_m_v ^ q._m_v)); }
  Quaternion	operator*(const Vector3d &v) const 
  { return Quaternion(-_m_v * v, _m_s * v + (_m_v ^ v)); }
  Quaternion	operator*(const Point3d &v) const 
  { return Quaternion(-_m_v * Vector3d(v),
		      Vector3d(v) * _m_s + (_m_v ^ Vector3d(v))); }
  Quaternion	operator+(const Quaternion &q) const 
  { return Quaternion(_m_s + q._m_s, _m_v + q._m_v); }
  Quaternion	operator-(const Quaternion &q) const 
  { return Quaternion(_m_s - q._m_s, _m_v - q._m_v); }
  double	norm2() const { return _m_s * _m_s + _m_v * _m_v; }
  double	norm() const { return sqrt(norm2()); }
  void		eqUnit() { operator*=(1.0 / norm()); }
  Quaternion	unit() const { return operator*(1.0 / norm()); }
  void		eqLog();
  Quaternion	log() const;
  Matrix4d	mlog() const;
  void		eqExp();
  Quaternion	exp() const;
  Quaternion	conjugate() const { return Quaternion(_m_s, -_m_v); }
  void		eqConjugate() { _m_v = -_m_v; }
  Quaternion	inverse() const { return conjugate() / norm2(); }
  void		eqSymx() { _m_v(1) = -_m_v(1); _m_v(2) = -_m_v(2); }
  void		eqSymy() { _m_v(2) = -_m_v(2); _m_v(0) = -_m_v(0); }
  void		eqSymz() { _m_v(0) = -_m_v(0); _m_v(1) = -_m_v(1); }
  void		eqInverse() { eqConjugate(); operator/(norm2()); }
  friend std::ostream &operator<<(std::ostream &os, const Quaternion &q)
  { return os << q._m_v << q._m_s; }
  friend std::istream &operator>>(std::istream &is, Quaternion &q)
  { return is >> q._m_v >> q._m_s; }
  double	&operator()(const unsigned i) { return i? _m_v(i - 1): _m_s; }
  double	operator()(const unsigned i) const
  { return i? _m_v(i - 1): _m_s; }
  void		glRotate() const;
  static Quaternion	creator(const double angle, const Vector3d &n);
  virtual Point3d	&operator()(Point3d &p) const;
  virtual Point3d	&operator()(Point3d &p, const double t) const;
  virtual void		operator()(Point3d &p, Vector3d &v) const;
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  double	det() const { return 1.0; }
};

// ************************************************************************* //
inline unsigned	Quaternion::nbSteps(const BBox3d	&b,
				    const double	k) const
{
  const Vector3d	v(_m_v.unit());
  Vector3d	p;
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      p = Vector3d(b.corner(i));
      l2tmp = (p - (p * v) * v).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * acos(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
inline void	Quaternion::cast(const Transformation &t)
{
  const Quaternion *q(dynamic_cast<const Quaternion *>(&t));

  if (!q)
    {
      std::cerr << "A Quaternion can only cast a Quaternion." << std::endl;
      std::exit(1);
    }
  *this = *q;
}

// ************************************************************************* //
inline Point3d	&Quaternion::operator()(Point3d &p) const
{
  p = Point3d((_m_s * _m_s - _m_v * _m_v) * Vector3d(p)
	      + (2.0 * (_m_v * Vector3d(p))) * _m_v
	      + (2.0 * _m_s) * (_m_v ^ Vector3d(p)));
  return p;
}

// ************************************************************************* //
inline Point3d	&Quaternion::operator()(Point3d		&p,
					const double	t) const
{
  const Quaternion	qt(operator^(t));
  
  qt(p);
  return p;
}

// ************************************************************************* //
inline void	Quaternion::operator()(Point3d	&p,
				       Vector3d	&v) const
{
  p = Point3d((_m_s * _m_s - _m_v * _m_v) * Vector3d(p)
	      + (2.0 * (_m_v * Vector3d(p))) * _m_v
	      + (2.0 * _m_s) * (_m_v ^ Vector3d(p)));
  v = v * (_m_s * _m_s - _m_v * _m_v)
    + _m_v * (2.0 * (_m_v * v))
    + (_m_v ^ v) * (2.0 * _m_s);
}

// ************************************************************************* //
/// DESCRIPTION: Constructor, for easy construction.
// ************************************************************************* //
inline Quaternion	Quaternion::creator(const double	angle,
					    const Vector3d	&n)
{
  const double	angle2(angle * 0.5);
  const double	nlen2(n.norm2());

  if (1.0e-8 < nlen2)
    return Quaternion(cos(angle2), n * (sin(angle2) / sqrt(nlen2)));
  return Quaternion(1.0, Vector3d(0.0));
}

// ************************************************************************* //
/// DESCRIPTION: this needs to be a unit quaternion.
// ************************************************************************* //
inline void	Quaternion::eqLog()
{
  const double	sinTheta(_m_v.norm());

  if (sinTheta == 0.0)
    _m_v = 0.0;
  else
    _m_v *= atan2(sinTheta, _m_s) / sinTheta;
  _m_s = 0.0;
}

// ************************************************************************* //
/// DESCRIPTION: this needs to be a unit quaternion.
// ************************************************************************* //
inline Quaternion	Quaternion::log() const
{
  const double	sinTheta(_m_v.norm());
    
  if (sinTheta == 0.0)
    return Quaternion(0.0);
  return Quaternion(0.0, (atan2(sinTheta, _m_s) / sinTheta) * _m_v);
}

// ************************************************************************* //
/// DESCRIPTION: Produces a unit quaternion.
// ************************************************************************* //
inline void	Quaternion::eqExp()
{
  const double	theta(_m_v.norm());

  if (theta == 0.0)
    _m_s = 1.0, _m_v = 0.0;
  else
    _m_s = cos(theta), _m_v *= sin(theta) / theta;
}

// ************************************************************************* //
/// DESCRIPTION: Produces a unit quaternion.
// ************************************************************************* //
inline Quaternion	Quaternion::exp() const
{
  const double	theta(_m_v.norm());

  if (theta == 0.0)
    return Quaternion(1.0, 0.0);
  return Quaternion(cos(theta), (sin(theta) / theta) * _m_v);
}

// ************************************************************************* //
inline void	Quaternion::glRotate() const
{
  glRotated(RAD2DEG(2.0 * myacos(_m_s)), _m_v.x, _m_v.y, _m_v.z);
}

// ************************************************************************* //
/// DESCRIPTION: this need to be a unit quaternion.
// ************************************************************************* //
inline Quaternion	Quaternion::operator^(const double d) const
{
  const double	sinTheta(_m_v.norm());
  
  if (sinTheta == 0.0)
    return Quaternion(1.0, 0.0);

  const double	angle((atan2(sinTheta, _m_s) * d));

  return Quaternion(cos(angle), _m_v * (sin(angle) / sinTheta));
}

// ************************************************************************* //
inline Quaternion	&Quaternion::operator^=(const double d)
{
  const double	sinTheta(_m_v.norm());
  
  if (sinTheta == 0.0)
    _m_s = 1.0, _m_v = 0.0;
  else
    {
      const double	angle((atan2(sinTheta, _m_s) * d));

      _m_s = cos(angle), _m_v *= (sin(angle) / sinTheta);
    }
  return *this;
}

// ************************************************************************* //
/// DESCRIPTION: Basic Matrix2i class.
// ************************************************************************* //
struct Matrix2i
{
  int	_m_m[4];
public:
  Matrix2i() {}
  int	&operator()(unsigned i, unsigned j) { return _m_m[2 * j + i]; }
  int	operator()(unsigned i, unsigned j) const {return _m_m[2 * j + i];}
  int	operator()(unsigned i) const {return _m_m[i];}
  void	glGet(const GLenum pname) { glGetIntegerv(pname, _m_m); }
  friend std::ostream	&operator<<(std::ostream &os, const Matrix2i &m)
  { for (unsigned i(0); i < 4; ++i) os << m._m_m[i] << " "; return os; }
  friend std::istream	&operator>>(std::istream &is, Matrix2i &m)
  { for (unsigned i(0); i < 4; ++i) is >> m._m_m[i]; return is; }
};

// ************************************************************************* //
/// DESCRIPTION: Basic Matrix3d class.
// ************************************************************************* //
class Matrix3d
{
  double	_m_m[9];
public:
  Matrix3d() {}
  Matrix3d(const Vector3d &v0, const Vector3d &v1, const Vector3d &v2);
  void		eqCommatrix();
  void		eqInverse();
  Vector3d	operator*(const Vector3d &v) const;
  Point3d	operator*(const Point3d &p) const;
  double	&operator()(unsigned i, unsigned j) { return _m_m[3 * j + i]; }
  double	operator()(unsigned i, unsigned j) const 
  { return _m_m[3 * j + i]; }
  virtual void	operator()(Point3d &p, Vector3d &v) const;
  virtual Point3d	&operator()(Point3d &p) const;
  //virtual Point3d	&operator()(Point3d &p, const double t) const;
  double	det() const;
  friend std::ostream	&operator<<(std::ostream &os, const Matrix3d &m)
  { for (unsigned i(0); i < 9; ++i) os << m._m_m[i] << " "; return os; }
};

// ************************************************************************* //
/// DESCRIPTION: this = this ^ -1
// ************************************************************************* //
inline void	Matrix3d::eqInverse()
{
  const double	_m_m0(_m_m[0]), _m_m1(_m_m[1]), _m_m2(_m_m[2]);
  const double	_m_m3(_m_m[3]), _m_m4(_m_m[4]), _m_m6(_m_m[6]);
  const double	invd(1.0 / (_m_m0 * (_m_m4 * _m_m[8] - _m_m[5] * _m_m[7])
			    + _m_m1 * (_m_m[5] * _m_m6 - _m_m3 * _m_m[8])
			    + _m_m2 * (_m_m3 * _m_m[7] - _m_m4 * _m_m6)));

  _m_m[0] = (_m_m4 * _m_m[8] - _m_m[5] * _m_m[7]) * invd;
  _m_m[1] = (_m_m[7] * _m_m2 - _m_m[8] * _m_m1) * invd;
  _m_m[2] = (_m_m1 * _m_m[5] - _m_m2 * _m_m4) * invd;
  _m_m[3] = (_m_m[5] * _m_m[6] - _m_m3 * _m_m[8]) * invd;
  _m_m[4] = (_m_m[8] * _m_m0 - _m_m[6] * _m_m2) * invd;
  _m_m[5] = (_m_m2 * _m_m3 - _m_m0 * _m_m[5]) * invd;
  _m_m[6] = (_m_m3 * _m_m[7] - _m_m4 * _m_m[6]) * invd;
  _m_m[7] = (_m_m6 * _m_m1 - _m_m[7] * _m_m0) * invd;
  _m_m[8] = (_m_m0 * _m_m4 - _m_m1 * _m_m3) * invd;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix determinant.
// ************************************************************************* //
inline double	Matrix3d::det() const
{
  return _m_m[0] * (_m_m[4] * _m_m[8] - _m_m[5] * _m_m[7])
    + _m_m[1] * (_m_m[5] * _m_m[6] - _m_m[3] * _m_m[8])
    + _m_m[2] * (_m_m[3] * _m_m[7] - _m_m[4] * _m_m[6]);
}

// ************************************************************************* //
inline void	Matrix3d::operator()(Point3d	&p,
				     Vector3d	&v) const
{
  p = operator*(p);
  v = operator*(v);
}

// ************************************************************************* //
inline Point3d	&Matrix3d::operator()(Point3d	&p) const
{
  p = operator*(p);
  return p;
}

// ************************************************************************* //
inline Matrix3d::Matrix3d(const Vector3d &v0,
			  const Vector3d &v1,
			  const Vector3d &v2)
{
  _m_m[0] = v0(0), _m_m[3] = v1(0), _m_m[6] = v2(0);
  _m_m[1] = v0(1), _m_m[4] = v1(1), _m_m[7] = v2(1);
  _m_m[2] = v0(2), _m_m[5] = v1(2), _m_m[8] = v2(2);
}

// ************************************************************************* //
/// DESCRIPTION: Matrix of the cofactors.
// ************************************************************************* //
inline void	Matrix3d::eqCommatrix()
{
  const double	_m_m0(_m_m[0]), _m_m3(_m_m[3]);
  const double	_m_m6(_m_m[6]);
  const double	_m_m1(_m_m[1]), _m_m4(_m_m[4]);
  const double	_m_m7(_m_m[7]);

  _m_m[0] = _m_m[4] * _m_m[8] - _m_m[5] * _m_m[7];
  _m_m[3] = _m_m[7] * _m_m[2] - _m_m[8] * _m_m[1];
  _m_m[6] = _m_m[1] * _m_m[5] - _m_m[2] * _m_m[4];
  _m_m[1] = _m_m[5] * _m_m6 - _m_m3 * _m_m[8];
  _m_m[4] = _m_m[8] * _m_m0 - _m_m6 * _m_m[2];
  _m_m[7] = _m_m[2] * _m_m3 - _m_m0 * _m_m[5];
  _m_m[2] = _m_m3 * _m_m7 - _m_m4 * _m_m6;
  _m_m[5] = _m_m6 * _m_m1 - _m_m7 * _m_m0;
  _m_m[8] = _m_m0 * _m_m4 - _m_m1 * _m_m3;
}

// ************************************************************************* //
inline Vector3d	Matrix3d::operator*(const Vector3d &v) const
{
  return Vector3d(_m_m[0] * v(0) + _m_m[3] * v(1) + _m_m[6] * v(2),
		  _m_m[1] * v(0) + _m_m[4] * v(1) + _m_m[7] * v(2),
		  _m_m[2] * v(0) + _m_m[5] * v(1) + _m_m[8] * v(2));
}

// ************************************************************************* //
inline Point3d	Matrix3d::operator*(const Point3d &p) const
{
  return Point3d(_m_m[0] * p(0) + _m_m[3] * p(1) + _m_m[6] * p(2),
		 _m_m[1] * p(0) + _m_m[4] * p(1) + _m_m[7] * p(2),
		 _m_m[2] * p(0) + _m_m[5] * p(1) + _m_m[8] * p(2));
}

// ************************************************************************* //
/// DESCRIPTION: Basic Matrix4d class, openGL compatible.
///  The most generic container for transformation. Not the 
///  most efficient. Use Quaternion Translation Rotation Scale and ScaleT for 
///  fast transformations.
// ************************************************************************* //
struct Matrix4d: public Transformation
{
  double	_m_m[16];
public:
  Matrix4d() {}
  Matrix4d(const Matrix4d &m);
  Matrix4d(const Vector3d &x, const Vector3d &y, const Vector3d &z,
	   const Point3d &o);
  Matrix4d(const double a);
  Matrix4d(double a00, double a01, double a02, double a03, 
	   double a10, double a11, double a12, double a13, 
	   double a20, double a21, double a22, double a23, 
	   double a30, double a31, double a32, double a33);
  Transformation	*copy() const { return new Matrix4d(*this); }
  explicit	Matrix4d(const Quaternion &q);
  explicit	Matrix4d(const Translation &t);
  explicit	Matrix4d(const Rotation &r);
  explicit	Matrix4d(const Scale &s);
  explicit	Matrix4d(const ScaleT &s);
  Matrix4d	asMatrix() const { return *this; }
  double	operator()(unsigned i, unsigned j) const 
  { return _m_m[4 * j + i]; }
  double	&operator()(unsigned i, unsigned j)
  { return _m_m[4 * j + i]; }
  friend std::ostream	&operator<<(std::ostream &os, const Matrix4d &m)
  { for (unsigned i(0); i < 16; ++i) os << m._m_m[i] << " "; return os; }
  friend std::istream	&operator>>(std::istream &is, Matrix4d &m)
  { for (unsigned i(0); i < 16; ++i) is >> m._m_m[i]; return is; }
  Matrix4d	inverse() const;
  Matrix4d	myInverse() const;
  void		eqInverse();
  void		eqTranspose();
  void		flipHandedness();
  void		eqSymx();
  void		eqSymy();
  void		eqSymz();
  Matrix4d	log() const;
  Matrix4d	mlog() const { return log(); }
  void		eqLog();
  Matrix4d	exp() const;
  void		eqExp();
  Matrix4d	sqrt() const;
  void		eqSqrt();
  double	normInfty() const;
  Matrix4d	operator*(const Matrix4d &m) const;
  Matrix4d	operator*(const double d) const;
  friend Matrix4d	operator*(const double d, const Matrix4d &m)
  { return m * d; }
  Matrix4d	&operator*=(const Matrix4d &that);
  Matrix4d	&operator*=(const double d);
  Matrix4d	&operator+=(const Matrix4d &that);
  Matrix4d	&operator-=(const Matrix4d &that);
  Matrix4d	operator+(const Matrix4d &that) const;
  Matrix4d	operator-(const Matrix4d &that) const;
  Matrix4d	operator^(const unsigned k) const;
  Matrix4d	operator^(const double t) const;
  Matrix4d	&operator^=(const double t);
  Matrix4d	&operator^=(const unsigned k);
  Vector3d	operator*(const Vector3d &v) const;
  Point3d	operator*(const Point3d &p) const;
  BBox3d	operator*(const BBox3d &b) const;
  double	det() const;
  static const Matrix4d	&identity();
  static const Matrix4d	&symx();
  static const Matrix4d	&symy();
  static const Matrix4d	&symz();
  static const Matrix4d	&sym(const unsigned sym);
  void	glGet(const GLenum pname) { glGetDoublev(pname, _m_m); }
  Vector3d	getx() const { return Vector3d(_m_m[0], _m_m[1], _m_m[2]); }
  Vector3d	gety() const { return Vector3d(_m_m[4], _m_m[5], _m_m[6]); }
  Vector3d	getz() const { return Vector3d(_m_m[8], _m_m[9], _m_m[10]); }
  Point3d	geto() const { return Point3d(_m_m[12], _m_m[13], _m_m[14]); }
  void		glMultMatrix() const { glMultMatrixd(_m_m); }
  virtual Point3d	&operator()(Point3d &p) const
  { p = operator*(p); return p; }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  { p = operator^(t) * p; return p; }
  virtual void		operator()(Point3d &p, Vector3d &v) const
  { p = operator*(p); v = operator*(v); }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
};

// ************************************************************************* //
/// DESCRIPTION: Determinant.
// ************************************************************************* //
inline double	Matrix4d::det() const
{
  Matrix3d	sub;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4);
      det += operator()(i, 0) * sub.det();
    }
  return det;
}

// ************************************************************************* //
inline unsigned	Matrix4d::nbSteps(const BBox3d	&b,
				  const double	k) const
{
  const Matrix4d	logThis(log());
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = Vector3d((logThis * b.corner(i))).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = ::sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
inline void	Matrix4d::cast(const Transformation &t)
{
  const Matrix4d *m(dynamic_cast<const Matrix4d *>(&t));

  if (!m)
    {
      std::cerr << "A Matrix4d can only cast a Matrix4d." << std::endl;
      std::exit(1);
    }
  *this = *m;
}

// ************************************************************************* //
inline Matrix4d	Transformation::operator*(const Matrix4d &m) const
{
  return asMatrix() * m;
}

// ************************************************************************* //
inline Matrix4d	Transformation::operator*(const Transformation &t) const
{
  return operator*(t.asMatrix());
}

// ************************************************************************* //
inline Matrix4d	Quaternion::asMatrix() const
{
  return Matrix4d(*this);
}

// ************************************************************************* //
inline Matrix4d	Quaternion::mlog() const
{
  const double	sinTheta(_m_v.norm());
    
  if (sinTheta == 0.0)
    return Matrix4d(0.0);

  const Vector3d	n((2.0 * atan2(sinTheta, _m_s) / sinTheta) * _m_v);

  return Matrix4d(0.0, -n.z, n.y, 0.0,
		  n.z, 0.0, -n.x, 0.0,
		  -n.y, n.x, 0.0, 0.0,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
inline Vector3d	&Vector3d::operator*=(const Matrix4d	&m)
{
  const double	tmpx(x), tmpy(y); 

  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z;
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z;
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z;
  return *this;
}

// ************************************************************************* //
inline Vector3d	&Vector3d::operator*=(const Matrix3d	&m)
{
  const double	tmpx(x), tmpy(y); 

  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z;
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z;
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z;
  return *this;
}

// ************************************************************************* //
inline Point3d	&Point3d::operator*=(const Matrix4d	&m)
{
  const double	tmpx(x), tmpy(y); 
  
  x = m(0, 0) * x + m(0, 1) * y + m(0, 2) * z + m(0, 3);
  y = m(1, 0) * tmpx + m(1, 1) * y + m(1, 2) * z + m(1, 3);
  z = m(2, 0) * tmpx + m(2, 1) * tmpy + m(2, 2) * z + m(2, 3);
  return *this;
}

// ************************************************************************* //
/// DESCRIPTION: Value constructor
// ************************************************************************* //
inline	Matrix4d::Matrix4d(const double a)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] = a;
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Vector3d	&x,
			  const Vector3d	&y,
			  const Vector3d	&z,
			  const Point3d		&o)
{
  _m_m[0] = x(0); _m_m[4] = y(0); _m_m[8] = z(0); _m_m[12] = o(0);
  _m_m[1] = x(1); _m_m[5] = y(1); _m_m[9] = z(1); _m_m[13] = o(1);
  _m_m[2] = x(2); _m_m[6] = y(2); _m_m[10] = z(2); _m_m[14] = o(2);
  _m_m[3] = 0.0; _m_m[7] = 0.0; _m_m[11] = 0.0; _m_m[15] = 1.0; 
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Matrix4d &m): Transformation()
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] = m._m_m[i];
}

// ************************************************************************* //
/// DESCRIPTION: Value constructor.
// ************************************************************************* //
inline	Matrix4d::Matrix4d(double a00, double a01, double a02, double a03, 
			   double a10, double a11, double a12, double a13, 
			   double a20, double a21, double a22, double a23, 
			   double a30, double a31, double a32, double a33)
{
  _m_m[0] = a00, _m_m[4] = a01, _m_m[8] = a02, _m_m[12] = a03;
  _m_m[1] = a10, _m_m[5] = a11, _m_m[9] = a12, _m_m[13] = a13;
  _m_m[2] = a20, _m_m[6] = a21, _m_m[10] = a22, _m_m[14] = a23;
  _m_m[3] = a30, _m_m[7] = a31, _m_m[11] = a32, _m_m[15] = a33;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix inverse which supports -O3 optimization.
// ************************************************************************* //
inline Matrix4d	Matrix4d::inverse() const
{
  Matrix4d	inv;
  Matrix3d	sub0, sub1;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4),
	    sub1(si, sj) = s * operator()((i + 1 + si) % 4, (3 + sj) % 4);
      inv(0, i) = sub0.det();
      inv(2, i) = sub1.det();
      det += operator()(i, 0) * sub0.det();
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = -s * operator()((i + 1 + si) % 4, (2 + sj) % 4),
	    sub1(si, sj) = -s * operator()((i + 1 + si) % 4, sj % 4);
      inv(1, i) = sub0.det();
      inv(3, i) = sub1.det();
    }
  if (fabs(det) < 1.0e-20)
    {
      std::cerr << "Singular matrix" << std::endl;
      std::exit(1);
    }
  return inv * (1.0 / det);
}

// ************************************************************************* //
/// DESCRIPTION: Matrix inversion.
// ************************************************************************* //
inline void	Matrix4d::eqInverse()
{
  Matrix4d	inv;
  Matrix3d	sub0, sub1;
  double	s, det = 0.0;

  for (unsigned i(0); i < 4; ++i)
    {
      s = (i % 2)? -1.0: 1.0;
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = s * operator()((i + 1 + si) % 4, (1 + sj) % 4),
	    sub1(si, sj) = s * operator()((i + 1 + si) % 4, (3 + sj) % 4);
      inv(0, i) = sub0.det();
      inv(2, i) = sub1.det();
      det += operator()(i, 0) * sub0.det();
      for (unsigned si(0); si < 3; ++si)
	for (unsigned sj(0); sj < 3; ++sj)
	  sub0(si, sj) = -s * operator()((i + 1 + si) % 4, (2 + sj) % 4),
	    sub1(si, sj) = -s * operator()((i + 1 + si) % 4, sj % 4);
      inv(1, i) = sub0.det();
      inv(3, i) = sub1.det();
    }
  if (det < 1.0e-20)
    {
      std::cerr << "Singular matrix" << std::endl;
      std::exit(1);
    }
  *this = inv * (1.0 / det);
}

// ************************************************************************* //
/// DESCRIPTION: Matrix transpose.
// ************************************************************************* //
inline void	Matrix4d::eqTranspose()
{
  double	tmp;

  tmp = _m_m[1], _m_m[1] = _m_m[4], _m_m[4] = tmp;
  tmp = _m_m[2], _m_m[2] = _m_m[8], _m_m[8] = tmp;
  tmp = _m_m[3], _m_m[3] = _m_m[12], _m_m[12] = tmp;
  tmp = _m_m[6], _m_m[6] = _m_m[9], _m_m[9] = tmp;
  tmp = _m_m[7], _m_m[7] = _m_m[13], _m_m[13] = tmp;
  tmp = _m_m[11], _m_m[11] = _m_m[14], _m_m[14] = tmp;
}

// ************************************************************************* //
inline void	Matrix4d::flipHandedness()
{
  operator()(2, 0) = -operator()(2, 0);
  operator()(2, 1) = -operator()(2, 1);
  operator()(2, 2) = -operator()(2, 2);
  operator()(2, 3) = -operator()(2, 3);
}

// ************************************************************************* //
inline void	Matrix4d::eqSymx()
{
  *this = Matrix4d::symx() * *this;
}

// ************************************************************************* //
inline void	Matrix4d::eqSymy()
{
  *this = Matrix4d::symy() * *this;
}

// ************************************************************************* //
inline void	Matrix4d::eqSymz()
{
  *this = Matrix4d::symz() * *this;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix logarithm.
// ************************************************************************* //
inline Matrix4d	Matrix4d::log() const
{
  double	i(1.0);
  Matrix4d	z, x , a(*this);
  unsigned	l(4096), k(0);
  
  while (0.5 < (a - identity()).normInfty())
    {
      a.eqSqrt();
      ++k;
    }
  z = x = a = Matrix4d::identity() - a;
  while (1.0e-6 < z.normInfty() && --l)
    {
      z *= a;
      x += z * (1.0 / ++i);
    }
  x *= -double(1 << k);
  return x;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix logarithm.
// ************************************************************************* //
inline void	Matrix4d::eqLog()
{
  double	i(1.0);
  Matrix4d	z, &x(*this), a(*this);
  unsigned	l(4096), k(0);
  
  while (0.5 < (a - identity()).normInfty())
    {
      a.eqSqrt();
      ++k;
    }
  z = x = a = Matrix4d::identity() - a;
  while (1.0e-6 < z.normInfty() && --l)
    {
      z *= a;
      x += z * (1.0 / ++i);
    }
  x *= -double(1 << k);
}

// ************************************************************************* //
/// DESCRIPTION: Matrix exponential.
// ************************************************************************* //
inline Matrix4d	Matrix4d::exp() const
{
  Matrix4d	d(identity()), n(identity());
  Matrix4d	x(identity()), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);

  a *= 1.0 / double(powj);
  x *= a;
  n += x * 0.5;
  d -= x * 0.5;
  x *= a;
  n += x * 0.113636363636;
  d += x * 0.113636363636;
  x *= a;
  n += x * 0.015151515151;
  d -= x * 0.015151515151;
  x *= a;
  n += x * 0.001262626262;
  d += x * 0.001262626262;
  x *= a;
  n += x * 0.000063131313;
  d -= x * 0.000063131313;

  d.eqInverse();
  x = d * n;

  for (unsigned i(1); i <= j; ++i)
    x *= x;
  return x;
  /*
  Matrix4d	d(identity()), n(identity());
  Matrix4d	x(identity()), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);
  double	c(1.0), k, q(6.0);

  a *= 1.0 / double(powj);
  for (k = 1.0; k < q; ++k)
    {
      c *= (q - k + 1.0) / (k * (2.0 * q - k + 1.0));
      x *= a;
      n += x * c;
      d += x * ((int(k) % 2)? -c: c);
    }
  d.eqInverse();
  x = d * n;
  x ^= powj;
  return x;
  */
}

// ************************************************************************* //
/// DESCRIPTION: Matrix exponential.
// ************************************************************************* //
inline void	Matrix4d::eqExp()
{
  Matrix4d	d(identity()), n(identity());
  Matrix4d	&x(*this), a(*this);
  unsigned	j(max(0, 1 + int(floor(::log(a.normInfty())))));
  unsigned	powj(1 << j);
  double	c(1.0), k, q(6.0);

  a *= 1.0 / double(powj);
  for (k = 1.0; k < q; ++k)
    {
      c *= (q - k + 1.0) / (k * (2.0 * q - k + 1.0));
      x *= a;
      n += x * c;
      d += x * ((int(k) % 2)? -c: c);
    }
  d.eqInverse();
  x = d * n;
  x ^= powj;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix square root.
// ************************************************************************* //
inline Matrix4d	Matrix4d::sqrt() const
{
  Matrix4d	x(*this), ix, y(identity()), iy;
  unsigned	l(4096);

  while (1.0e-6 < (x * x - *this).normInfty() && --l)
    {
      ix = x.inverse();
      iy = y.inverse();
      x += iy;
      x *= 0.5;
      y += ix;
      y *= 0.5;
    }
  return x;
}

// ************************************************************************* //
/// DESCRIPTION: Matrix square root.
// ************************************************************************* //
inline void	Matrix4d::eqSqrt()
{
  Matrix4d	&x(*this), ix, y(identity()), iy;
  unsigned	l(4096);

  while (1.0e-6 < (x * x - *this).normInfty() && --l)
    {
      ix = x.inverse();
      iy = y.inverse();
      x += iy;
      x *= 0.5;
      y += ix;
      y *= 0.5;
    }
}

// ************************************************************************* //
/// DESCRIPTION: Matrix norm.
// ************************************************************************* //
inline double	Matrix4d::normInfty() const
{
  unsigned	i, j;
  double	norm(0.0), tmp;
  
  for (i = 0; i < 4; ++i)
    {
      for (tmp = 0.0, j = 0; j < 4; ++j)
	tmp += fabs(operator()(i, j));
      if (norm < tmp)
	norm = tmp;
    }	
  return norm;
}

// ************************************************************************* //
inline Matrix4d	Matrix4d::operator*(const Matrix4d &that) const
{
  Matrix4d	ret(0.0);

  for (unsigned i(0); i < 4; ++i)
    for (unsigned j(0); j < 4; ++j)
      for (unsigned k(0); k < 4; ++k)
	ret(i, j) += operator()(i, k) * that(k, j);
  return ret;
}

// ************************************************************************* //
/// DESCRIPTION: this *= that => this = this * that
// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator*=(const Matrix4d &that)
{
  Matrix4d	ret(0.0);

  for (unsigned i(0); i < 4; ++i)
    for (unsigned j(0); j < 4; ++j)
      for (unsigned k(0); k < 4; ++k)
	ret(i, j) += operator()(i, k) * that(k, j);
  *this = ret;
  return *this;
}

// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator*=(const double d)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] *= d;
  return *this;
}

// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator+=(const Matrix4d &that)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] += that._m_m[i];
  return *this;
}

// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator-=(const Matrix4d &that)
{
  for (unsigned i(0); i < 16; ++i)
    _m_m[i] -= that._m_m[i];
  return *this;
}

// ************************************************************************* //
inline Matrix4d	Matrix4d::operator*(const double d) const
{
  return Matrix4d
    (_m_m[0] * d, _m_m[4] * d, _m_m[8] * d, _m_m[12] * d,
     _m_m[1] * d, _m_m[5] * d, _m_m[9] * d, _m_m[13] * d,
     _m_m[2] * d, _m_m[6] * d, _m_m[10] * d, _m_m[14] * d,
     _m_m[3] * d, _m_m[7] * d, _m_m[11] * d, _m_m[15] * d);
}

// ************************************************************************* //
inline Matrix4d	Matrix4d::operator+(const Matrix4d &that) const
{
  return Matrix4d
    (_m_m[0] + that._m_m[0], _m_m[4] + that._m_m[4],
     _m_m[8] + that._m_m[8], _m_m[12] + that._m_m[12],
     _m_m[1] + that._m_m[1], _m_m[5] + that._m_m[5],
     _m_m[9] + that._m_m[9], _m_m[13] + that._m_m[13],
     _m_m[2] + that._m_m[2], _m_m[6] + that._m_m[6],
     _m_m[10] + that._m_m[10], _m_m[14] + that._m_m[14],
     _m_m[3] + that._m_m[3], _m_m[7] + that._m_m[7],
     _m_m[11] + that._m_m[11], _m_m[15] + that._m_m[15]);
}

// ************************************************************************* //
inline Matrix4d	Matrix4d::operator-(const Matrix4d &that) const
{
  return Matrix4d
    (_m_m[0] - that._m_m[0], _m_m[4] - that._m_m[4],
     _m_m[8] - that._m_m[8], _m_m[12] - that._m_m[12],
     _m_m[1] - that._m_m[1], _m_m[5] - that._m_m[5],
     _m_m[9] - that._m_m[9], _m_m[13] - that._m_m[13],
     _m_m[2] - that._m_m[2], _m_m[6] - that._m_m[6],
     _m_m[10] - that._m_m[10], _m_m[14] - that._m_m[14],
     _m_m[13] - that._m_m[3], _m_m[7] - that._m_m[7],
     _m_m[11] - that._m_m[11], _m_m[15] - that._m_m[15]);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the k^th integer power of d.
// ************************************************************************* //
inline Matrix4d	Matrix4d::operator^(const unsigned k) const
{
  Matrix4d	ret(identity()), r(*this);

  for (unsigned i(1); i <= k; i *= 2, r *= r)
    if (k & i)
      ret *= r;
  return ret;
}

// ************************************************************************* //
inline Matrix4d	Matrix4d::operator^(const double t) const
{
  return (log() * t).exp();
}

// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator^=(const double t)
{
  eqLog();
  operator*=(t);
  eqExp();
  return *this;
}

// ************************************************************************* //
inline Matrix4d	&Matrix4d::operator^=(const unsigned k)
{
  Matrix4d	r(*this);

  *this = identity();
  for (unsigned i(1); i <= k; i *= 2, r *= r)
    if (k & i)
      operator*=(r);
  return *this;
}

// ************************************************************************* //
inline Vector3d	Matrix4d::operator*(const Vector3d &v) const
{
  return Vector3d(_m_m[0] * v.x + _m_m[4] * v.y + _m_m[8] * v.z,
		  _m_m[1] * v.x + _m_m[5] * v.y + _m_m[9] * v.z,
		  _m_m[2] * v.x + _m_m[6] * v.y + _m_m[10] * v.z);
}

// ************************************************************************* //
inline Point3d	Matrix4d::operator*(const Point3d &p) const
{
  return Point3d
    (_m_m[0] * p.x + _m_m[4] * p.y + _m_m[8] * p.z + _m_m[12],
     _m_m[1] * p.x + _m_m[5] * p.y + _m_m[9] * p.z + _m_m[13],
     _m_m[2] * p.x + _m_m[6] * p.y + _m_m[10] * p.z + _m_m[14]);
}

// ************************************************************************* //
inline BBox3d	Matrix4d::operator*(const BBox3d	&b) const
{
  BBox3d	ret(operator*(b.pmin));
  
  ret.fill(operator*(b.pmax));
  ret.fill(operator*(Point3d(b.pmax.x, b.pmin.y, b.pmin.z)));
  ret.fill(operator*(Point3d(b.pmin.x, b.pmax.y, b.pmin.z)));
  ret.fill(operator*(Point3d(b.pmin.x, b.pmin.y, b.pmax.z)));
  ret.fill(operator*(Point3d(b.pmin.x, b.pmax.y, b.pmax.z)));
  ret.fill(operator*(Point3d(b.pmax.x, b.pmin.y, b.pmax.z)));
  ret.fill(operator*(Point3d(b.pmax.x, b.pmax.y, b.pmin.z)));
  return ret;
}

// ************************************************************************* //
inline BBox3d	&BBox3d::operator*=(const Matrix4d &m)
{
  BBox3d	ret(m * pmin);
  
  ret.fill(m * pmax);
  ret.fill(m * Point3d(pmax.x, pmin.y, pmin.z));
  ret.fill(m * Point3d(pmin.x, pmax.y, pmin.z));
  ret.fill(m * Point3d(pmin.x, pmin.y, pmax.z));
  ret.fill(m * Point3d(pmin.x, pmax.y, pmax.z));
  ret.fill(m * Point3d(pmax.x, pmin.y, pmax.z));
  ret.fill(m * Point3d(pmax.x, pmax.y, pmin.z));
  *this = ret;
  return *this;
}

// ************************************************************************* //
inline const Matrix4d	&Matrix4d::identity()
{
  static const Matrix4d	id(1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);

  return id;
}

// ************************************************************************* //
inline const Matrix4d	&Matrix4d::symx()
{
  static const Matrix4d ix(-1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return ix;
}

// ************************************************************************* //
inline const Matrix4d	&Matrix4d::symy()
{
  static const Matrix4d iy(1.0, 0.0, 0.0, 0.0,
			   0.0, -1.0, 0.0, 0.0,
			   0.0, 0.0, 1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return iy;
}

// ************************************************************************* //
inline const Matrix4d	&Matrix4d::symz()
{
  static const Matrix4d iz(1.0, 0.0, 0.0, 0.0,
			   0.0, 1.0, 0.0, 0.0,
			   0.0, 0.0, -1.0, 0.0,
			   0.0, 0.0, 0.0, 1.0);
  return iz;
}

// ************************************************************************* //
inline const Matrix4d	&Matrix4d::sym(const unsigned sym)
{
  if (sym == 1)
    return symx();
  else if (sym == 2)
    return symy();
  else if (sym == 3)
    return symz();
  return identity();
}

// ************************************************************************* //
/// DESCRIPTION: Translation.
// ************************************************************************* //
struct Translation: public Transformation
{
  friend class	Matrix4d;
  Vector3d	_m_v;
public:
  Translation(const Vector3d &v): _m_v(v) {}
  Matrix4d	asMatrix() const { return Matrix4d(*this); }
  Translation	operator^(const double t) const
  { return Translation(_m_v * t); }
  Translation	&operator^=(const double t) { _m_v *= t; return *this; }
  Transformation	*copy() const { return new Translation(*this); }
  virtual Point3d	&operator()(Point3d &p) const
  { p += _m_v;  return p; }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  { p += _m_v * t; return p; }
  virtual void		operator()(Point3d &p, Vector3d &) const { p += _m_v;}
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_v(0) = -_m_v(0); }
  void		eqSymy() { _m_v(1) = -_m_v(1); }
  void		eqSymz() { _m_v(2) = -_m_v(2); }
  void		eqInverse() { _m_v = -_m_v; }
  double	det() const { return 1.0; }
  Matrix4d	mlog() const;
};

// ************************************************************************* //
inline unsigned	Translation::nbSteps(const BBox3d	&,
				     const double	k) const
{
  const double	s(ceil(1.875 / k * _m_v.norm()));
  
  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
inline Matrix4d	Translation::mlog() const
{
  return Matrix4d(0.0, 0.0, 0.0, _m_v.x,
		  0.0, 0.0, 0.0, _m_v.y,
		  0.0, 0.0, 0.0, _m_v.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
inline void	Translation::cast(const Transformation &t)
{
  const Translation *s(dynamic_cast<const Translation *>(&t));

  if (!s)
    {
      std::cerr << "A Translation only cast a Translation." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ************************************************************************* //
/// DESCRIPTION: Rotation around a pivot point. Uses class Quaternion.
// ************************************************************************* //
class Rotation: public Transformation
{
  friend class	Matrix4d;
protected:
  Quaternion	_m_q;
  Vector3d	_m_c;
public:
  Rotation(const Quaternion &q, const Point3d &c): _m_q(q), _m_c(c) {}
  Rotation(const Quaternion &q, const Vector3d &c): _m_q(q), _m_c(c) {}
  Matrix4d	asMatrix() const { return Matrix4d(*this); }
  Rotation	operator^(const double t) const
  { return Rotation(_m_q ^ t, _m_c); }
  Rotation		&operator^=(const double t) { _m_q ^= t; return *this;}
  virtual Transformation*copy() const { return new Rotation(_m_q, _m_c); }
  virtual Point3d	&operator()(Point3d &p) const
  { p -= _m_c; _m_q(p); p += _m_c; return p; }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  { p -= _m_c; _m_q(p, t); p += _m_c; return p; }
  virtual void		operator()(Point3d &p, Vector3d &v) const
  { p -= _m_c; _m_q(p, v); p += _m_c; }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); _m_q.eqSymx(); }
  void		eqSymy() { _m_c(1) = -_m_c(1); _m_q.eqSymy(); }
  void		eqSymz() { _m_c(2) = -_m_c(2); _m_q.eqSymz(); }
  void		eqInverse() { _m_q.eqInverse(); }
  double	det() const { return 1.0; }
  Matrix4d	mlog() const;
};

// ************************************************************************* //
/// DESCRIPTION: Transformation that is function of $x$.
// ************************************************************************* //
class UnitRotation: public Rotation
{
public:
  UnitRotation(const Quaternion &q, const Point3d &c): Rotation(q, c) {}
  UnitRotation(const Quaternion &q, const Vector3d &c): Rotation(q, c) {}

  virtual Transformation*copy() const { return new UnitRotation(_m_q, _m_c); }
  virtual Point3d	&operator()(Point3d &p) const
  {
    const Matrix4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vector3d(m * (p - _m_c)).norm()));
    
    p -= _m_c; q(p); p += _m_c; return p;
  }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  {
    const Matrix4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vector3d(m * (p - _m_c)).norm()));

    p -= _m_c; q(p, t); p += _m_c; return p;
  }
  virtual void		operator()(Point3d &p, Vector3d &v) const
  {
    const Matrix4d	m(*this);
    const Quaternion	q(_m_q ^ (1.0 / Vector3d(m * (p - _m_c)).norm()));

    p -= _m_c; q(p, v); p += _m_c;
  }
};

// ************************************************************************* //
inline unsigned	Rotation::nbSteps(const BBox3d	&b,
				  const double	k) const
{
  const Vector3d	v(_m_q._m_v.unit());
  Vector3d	p;
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      p = b.corner(i) - Point3d(_m_c);
      l2tmp = (p - (p * v) * v).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * acos(_m_q._m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
inline Matrix4d	Rotation::mlog() const
{
  const double	sinTheta(_m_q._m_v.norm());
    
  if (sinTheta == 0.0)
    return Matrix4d(0.0);

  const Vector3d	n((2.0 * atan2(sinTheta, _m_q._m_s) /
			   sinTheta) * _m_q._m_v);
  const Vector3d	cross(_m_c ^ n);

  return Matrix4d(0.0, -n.z, n.y, cross.x,
		  n.z, 0.0, -n.x, cross.y,
		  -n.y, n.x, 0.0, cross.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
inline void	Rotation::cast(const Transformation &t)
{
  const Rotation *r(dynamic_cast<const Rotation *>(&t));

  if (!r)
    {
      std::cerr << "A Rotation can only cast a Rotation." << std::endl;
      std::exit(1);
    }
  *this = *r;
}

// ************************************************************************* //
/// DESCRIPTION: Scale around center point.
// ************************************************************************* //
class Scale: public Transformation
{
  friend class	Matrix4d;
  double	_m_s;
  Point3d	_m_c;
public:
  Scale(const double s, const Point3d &c): _m_s(s), _m_c(c) {}
  Matrix4d	asMatrix() const { return Matrix4d(*this); }
  Scale		operator^(const double t) const
  { return Scale(pow(_m_s, t), _m_c); }
  Scale	&operator^=(const double t) { _m_s = pow(_m_s, t); return *this; }
  Transformation	*copy() const { return new Scale(_m_s, _m_c); }
  virtual Point3d	&operator()(Point3d &p) const
  { p = _m_c + (p - _m_c) * _m_s; return p; }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  { p = _m_c + (p - _m_c) * pow(_m_s, t); return p; }
  virtual void		operator()(Point3d &p, Vector3d &) const
  { p = _m_c + (p - _m_c) * _m_s; }
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); }
  void		eqSymy() { _m_c(1) = -_m_c(1); }
  void		eqSymz() { _m_c(2) = -_m_c(2); }
  void		eqInverse() { _m_s = 1.0 / _m_s; }
  double	det() const { return _m_s * _m_s * _m_s; }
  Matrix4d	mlog() const;
};

// ************************************************************************* //
inline unsigned	Scale::nbSteps(const BBox3d	&b,
			       const double	k) const
{
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = (b.corner(i) - Point3d(_m_c)).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * _m_s * log(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
inline Matrix4d	Scale::mlog() const
{
  const double logs(log(_m_s));

  return Matrix4d(logs, 0.0, 0.0, -_m_c.x * logs,
		  0.0, logs, 0.0, -_m_c.y * logs,
		  0.0, 0.0, logs, -_m_c.z * logs,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
inline void	Scale::cast(const Transformation &t)
{
  const Scale *s(dynamic_cast<const Scale *>(&t));

  if (!s)
    {
      std::cerr << "A Scale can only cast a Scale." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ************************************************************************* //
/// DESCRIPTION: Scale around center point, along axis _m_d only.
// ************************************************************************* //
class ScaleT: public Transformation
{
  friend class	Matrix4d;
  double	_m_s;
  Point3d	_m_c;
  Vector3d	_m_d;
public:
  ScaleT(const double s, const Point3d &c, const Vector3d &d):
    _m_s(s), _m_c(c), _m_d(d.unit()) {}
  Transformation	*copy() const { return new ScaleT(_m_s, _m_c, _m_d); }
  Matrix4d	asMatrix() const { return Matrix4d(*this); }
  ScaleT	operator^(const double t) const
  { return ScaleT(pow(_m_s, t), _m_c, _m_d); }
  ScaleT	&operator^=(const double t)
  { _m_s = pow(_m_s, t); return *this; }
  virtual Point3d	&operator()(Point3d &p) const
  { p += (_m_s - 1.0) * ((p - _m_c) * _m_d) * _m_d; return p; }
  virtual Point3d	&operator()(Point3d &p, const double t) const
  { p += (pow(_m_s, t) - 1.0) * ((p - _m_c) * _m_d) * _m_d; return p; }
  virtual void		operator()(Point3d &p, Vector3d &v) const;
  void		cast(const Transformation &t);
  unsigned	nbSteps(const BBox3d &b, const double k) const;
  void		eqSymx() { _m_c(0) = -_m_c(0); _m_d(0) = -_m_d(0); }
  void		eqSymy() { _m_c(1) = -_m_c(1); _m_d(1) = -_m_d(1);  }
  void		eqSymz() { _m_c(2) = -_m_c(2); _m_d(2) = -_m_d(2);  }
  void		eqInverse() { _m_s = 1.0 / _m_s; }
  double	det() const { return _m_s; }
  Matrix4d	mlog() const;
};

// ************************************************************************* //
inline unsigned	ScaleT::nbSteps(const BBox3d	&b,
				const double	k) const
{
  double	l2tmp, lmax(0.0);

  for (int i(0); i < 8; ++i)
    {
      l2tmp = (b.corner(i) - Point3d(_m_c)).norm2();
      if (lmax < l2tmp)
	lmax = l2tmp;
    }
  lmax = sqrt(lmax);

  const double s = ceil(1.875 / k * 2.0 * _m_s * log(_m_s) * lmax);

  if (s <= 1.0)
    return 1;
  return unsigned(s);
}

// ************************************************************************* //
/// DESCRIPTION: Logarithm of the transformation matrix.
// ************************************************************************* //
inline Matrix4d	ScaleT::mlog() const
{
  const double logs(log(_m_s));
  const double logsx(logs * _m_d.x), logsy(logs * _m_d.y),logsz(logs * _m_d.z);
  const double scal(Vector3d(_m_c) * _m_d);

  return 
    Matrix4d(logsx * _m_d.x, logsx * _m_d.y, logsx * _m_d.z, -logsx * scal,
	     logsy * _m_d.x, logsy * _m_d.y, logsy * _m_d.z, -logsy * scal,
	     logsz * _m_d.x, logsz * _m_d.y, logsz * _m_d.z, -logsz * scal,
	     0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
inline void	ScaleT::cast(const Transformation &t)
{
  const ScaleT *s(dynamic_cast<const ScaleT *>(&t));

  if (!s)
    {
      std::cerr << "A ScaleT can only cast a ScaleT." << std::endl;
      std::exit(1);
    }
  *this = *s;
}

// ************************************************************************* //
inline void	ScaleT::operator()(Point3d	&p,
				   Vector3d	&v) const
{
  p += (_m_s - 1.0) * ((p - _m_c) * _m_d) * _m_d;
  v += (_m_s - 1.0) * (v - (v * _m_d) * _m_d);
  v.eqUnit();
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Quaternion &q)
{
  const double	s(2.0 / q.norm());
  const double	xs(q(1) * s);
  const double	ys(q(2) * s);
  const double	zs(q(3) * s);
  const double	wx(q(0) * xs);
  const double	wy(q(0) * ys);
  const double	wz(q(0) * zs);
  const double	xx(q(1) * xs);
  const double	xy(q(1) * ys);
  const double	xz(q(1) * zs);
  const double	yy(q(2) * ys);
  const double	yz(q(2) * zs);
  const double	zz(q(3) * zs);

  *this = Matrix4d(1.0 - (yy + zz), xy - wz, xz + wy, 0.0,
		   xy + wz, 1.0 - (xx + zz), yz - wx, 0.0,
		   xz - wy, yz + wx, 1.0 - (xx + yy), 0.0,
		   0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Translation &t)
{
  const Vector3d	&v(t._m_v);

  *this = Matrix4d(1.0, 0.0, 0.0, v.x,
		   0.0, 1.0, 0.0, v.y,
		   0.0, 0.0, 1.0, v.z,
		   0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Rotation &r)
{
  const Quaternion	&q(r._m_q);
  const Point3d		c(r._m_c);

  const double	xs(q(1) * 2.0), ys(q(2) * 2.0), zs(q(3) * 2.0);
  const double	wx(q(0) * xs), wy(q(0) * ys), wz(q(0) * zs);
  const double	xx(q(1) * xs), xy(q(1) * ys), xz(q(1) * zs);
  const double	yy(q(2) * ys), yz(q(2) * zs), zz(q(3) * zs);
  const Matrix4d	mat(-yy - zz, xy - wz, xz + wy, 0.0, 
			    xy + wz, -xx - zz, yz - wx, 0.0, 
			    xz - wy, yz + wx, -xx - yy, 0.0,
			    0.0, 0.0, 0.0, 0.0);
  const Point3d		tmp = mat * c;

  *this = Matrix4d(1.0 + mat(0,0), mat(0, 1), mat(0, 2), -tmp(0),
		   mat(1,0), 1.0 + mat(1, 1), mat(1, 2), -tmp(1),
		   mat(2,0), mat(2, 1), 1.0 + mat(2, 2), -tmp(2),
		   0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const Scale &scl)
{
  const double	s(scl._m_s);
  const Point3d	&c(scl._m_c);

  *this = Matrix4d(s, 0.0, 0.0, (1.0 - s) * c(0),
		   0.0, s, 0.0, (1.0 - s) * c(1),
		   0.0, 0.0, s, (1.0 - s) * c(2),
		   0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
inline Matrix4d::Matrix4d(const ScaleT	&scl)
{
  const double		s(scl._m_s);
  const Point3d		&c(scl._m_c);
  const Vector3d	&v(scl._m_d);
  Vector3d		tmp0, tmp1;

  tmp0 = ((1.0 - s) * (Vector3d(c) * v)) * v;
  tmp1 = -(1.0 - s) * v;
  *this = Matrix4d
    (1.0 + tmp1(0) * v(0), tmp1(1) * v(0), tmp1(2) * v(0), tmp0(0),
     tmp1(0) * v(1), 1.0 + tmp1(1) * v(1), tmp1(2) * v(1), tmp0(1),
     tmp1(0) * v(2), tmp1(1) * v(2), 1.0 + tmp1(2) * v(2), tmp0(2),
     0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
#endif//__SWPTYPES_HH__

