/* ************************************************************************* **
** Alexis Angelidis							     **
** Started Mon Jan 28 12:24:46 MET 2002					     **
** ************************************************************************* */
#include <stdio.h>
#include <math.h>
#include <fstream>
#include "glTrackball.h"

/* ************************************************************************* */
#ifndef WIN32
namespace gl
{
#endif

/* ************************************************************************* */
#ifndef RAD2DEG
#define RAD2DEG(a) ((a) * 57.2957795130823208768)
#endif//RAD2DEG
#ifndef DEG2RAD
#define DEG2RAD(a) ((a) * 0.01745329251994329576)
#endif//DEG2RAD

/* ************************************************************************* */
GLuint	TrackBall::_s_list(0);

/* ************************************************************************* **
** Description: Constructor.						     **
**   rho is rotation angle around viewing direction.			     **
**   prox is distance from viewpoint to center point.			     **
**   rad1 & rad2 are the radius of the two drawn circles for graphic mode.   **
**   org is the position of center point.				     **
**   rot is orientation of center point coord set (relatively to world cs).  **
** ************************************************************************* */
TrackBall::TrackBall(View	&v): 
  _m_rot(1.0, 0.0), _m_org(0.0), _m_rho(0.0), _m_prox(10.0),
  _m_rad1(0.55), _m_rad2(0.85), _m_zNear(0.01), _m_zFar(100.0), _m_v(&v)
{
  resize();
}

/* ************************************************************************* */
void	TrackBall::reset()
{
  _m_rho = 0.0;
  _m_prox = 10.0;
  _m_rot = Quaternion(1.0, 0.0);
  _m_org = Point3d(0.0);
  _m_zNear = 0.01;
  _m_zFar = 100.0;
  resize();
}

/* ************************************************************************* */
void	TrackBall::resize()
{
  _m_factw = min(1.0, _m_v->_m_view(2) / double(_m_v->_m_view(3)));
  _m_facth = min(1.0, _m_v->_m_view(3) / double(_m_v->_m_view(2)));
  _m_rad = min(_m_factw, _m_facth);
}

// ************************************************************************* //
/// DESCRIPTION: Places the camera for renderman.
// ************************************************************************* //
void	TrackBall::riPlace(std::ofstream &ofs) const
{
  Matrix4d	p(_m_v->proj());
  Matrix4d	m(_m_v->model());

  m.flipHandedness();
  ofs << "ConcatTransform [" << m << "]" << std::endl;
}

/* ************************************************************************* **
**  _m_rad ... sphere radius in screen.					     **
**  _m_factw .. width frustrum coeff.					     **
**  _m_facth .. height frustrum coeff.					     **
** ************************************************************************* */
void	TrackBall::glRenderPlace() const
{
  static const GLfloat	lPosition0[] = {-0.5f, 0.5f, -1.0f, 0.0f};
  static const GLfloat	lPosition1[] = {1.2f, 0.0f, 1.0f, 0.0f};
  static const GLfloat	lPosition2[] = {-1.2f, 0.0f, 1.0f, 0.0f};
  static const GLfloat	lPosition3[] = {0.0f, 0.0f, 1.0f, 0.0f};

  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(50.0, _m_v->_m_view(2) / double(_m_v->_m_view(3)), _m_zNear,
		 _m_zFar);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glLightfv(GL_LIGHT0, GL_POSITION, lPosition0);
  glLightfv(GL_LIGHT1, GL_POSITION, lPosition1);
  glLightfv(GL_LIGHT2, GL_POSITION, lPosition2);
  glLightfv(GL_LIGHT3, GL_POSITION, lPosition3);
  glTranslated(0.0, 0.0, -_m_prox);
  glRotated(-_m_rho, 0.0, 0.0, 1.0);
  (-_m_rot).glRotate();
  (-_m_org).glTranslate();
  glPopAttrib();// GL_TRANSFORM_BIT
  _m_v->get();
}

// ************************************************************************* //
/// DESCRIPTION: Places the trackball for selection.
// ************************************************************************* //
void	TrackBall::glSelectPlace() const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_PROJECTION);
  gluPerspective(50.0, _m_v->_m_view(2) / double(_m_v->_m_view(3)), 
		 _m_zNear, _m_zFar);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslated(0.0, 0.0, -_m_prox);
  glRotated(-_m_rho, 0.0, 0.0, 1.0);
  (-_m_rot).glRotate();
  (-_m_org).glTranslate();
  glPopAttrib();// GL_TRANSFORM_BIT
  _m_v->get();
}

/* ************************************************************************* */
void	TrackBall::glRender() const
{
  glInitLists();
  displayCoordSet();
  displayTarget();
}

/* ************************************************************************* */
void	TrackBall::displayCoordSet() const
{
  const GLuint		tmp(_s_list + 1);
  const GLdouble	tmpScale = (_m_zFar - _m_zNear) / 500.0;

  glPushAttrib(GL_TRANSFORM_BIT);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(0.0, 0.0, -_m_prox);
  glColor3d(0.25, 1.0, 0.66);
  glScaled(tmpScale, tmpScale, tmpScale);
  glCallLists(1, GL_UNSIGNED_INT, &tmp);
  glPopMatrix();
  glPopAttrib();// GL_TRANSFORM_BIT
}

/* ************************************************************************* */
void	TrackBall::displayTarget() const
{
  glPushAttrib(GL_TRANSFORM_BIT);
  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(-_m_factw, _m_factw, -_m_facth, _m_facth, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();

  glLoadIdentity(); 

  glScaled(_m_rad, _m_rad, _m_rad);
  glPushMatrix();
  (_m_winRad < _m_rad2)? glColor3d(1.0, 0.0, 0.0): glColor3d(0.25, 1.0, 0.66);
  glScaled(_m_rad1, _m_rad1, _m_rad1);
  glCallLists(1, GL_UNSIGNED_INT, &_s_list);
  glPopMatrix();
  glPushMatrix();
  (_m_rad1 < _m_winRad && _m_winRad < _m_rad2)?
    glColor3d(1.0, 0.0, 0.0): glColor3d(0.25, 1.0, 0.66);
  glScaled(_m_rad2, _m_rad2, _m_rad2);
  glCallLists(1, GL_UNSIGNED_INT, &_s_list);
  glPopMatrix();

  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glPopAttrib();// GL_LIGHTING_BIT
  glPopAttrib();// GL_TRANSFORM_BIT
}

/* ************************************************************************* */
void	TrackBall::mousePressEvent(QMouseEvent	*e1)
{
  Point2d	v1(screen2normal(Point2d(e1->x(), e1->y())));

  _m_winRad = v1.norm();
}

/* ************************************************************************* */
void	TrackBall::mouseMove(const QMouseEvent	&e0,
			     const QMouseEvent	&e1)
{
  Vector2d	v0(screen2normal(Point2d(e0.x(), e0.y())));
  Vector2d	v1(screen2normal(Point2d(e1.x(), e1.y())));
  GLdouble	sphereRad;
  Point3d	eye(_m_v->_m_view(2) * 0.5, _m_v->_m_view(3) * 0.5, 0.0), p;
  Point3d	w0(e0.x(), _m_v->_m_view(3) * 0.5 - e0.y(), 10.0);
  Point3d	w1(e1.x(), _m_v->_m_view(3) * 0.5 - e1.y(), 10.0);
  Vector3d	n0, n1, u;

  eye = _m_v->unProject(eye);
  _m_winRad = v1.norm();
  _m_zNear = (eye - _m_org).norm() / 100.0;
  _m_zFar = 10.0 * (eye - _m_org).norm();
  switch (e1.state())
    {
    case Qt::LeftButton:
      if (_m_rad2 < _m_winRad)
	_m_prox -= (v1(1) - v0(1))  * (eye - _m_org).norm() * 0.5;
      else
	if (_m_rad1 < _m_winRad)
	  {
	    v0.eqUnit();
	    v1.eqUnit();
	    _m_rho -= RAD2DEG(v0(1) *  v1(0) - v0(0) * v1(1));
	  }
	else
	  {
	    p = normal2screen(Point3d(_m_rad1, 0.0, 0.0));
	    p = _m_v->unProject(p);
	    w0 = _m_v->unProject(w0);
	    w1 = _m_v->unProject(w1);
	    u = (p - eye).unit();
	    p = eye + ((_m_org - eye) * u) * u;
	    sphereRad = sqrt((eye - _m_org).norm2() - (eye - p).norm2());
	    n0 = (w0 - eye).unit();
	    n1 = (w1 - eye).unit();
	    w0.eqSphereIntersection(eye, n0, _m_org, sphereRad);
	    w1.eqSphereIntersection(eye, n1, _m_org, sphereRad);
	    n0 = (w0 - _m_org).unit();
	    n1 = (w1 - _m_org).unit();
	    _m_rot =_m_rot *
	      Quaternion::creator(10.0 * myacos(n1 * n0), (n1 ^ n0));
	    _m_rot.eqUnit();
	  }
      break;
    case Qt::MidButton:
      if (_m_winRad < _m_rad2)
	{
	  w0 = _m_v->unProject(w0);
	  w1 = _m_v->unProject(w1);
	  n0 = w0 - eye;
	  w0 = eye - ((eye - _m_org).norm2() / (n0 * (eye - _m_org))) * n0;
	  n1 = w1 - eye;
	  w1 = eye - ((eye - _m_org).norm2() / (n1 * (eye - _m_org))) * n1;
	  _m_org -= 8.0 * (w1 - w0);
	}
      else
	_m_org -= (v1(1) - v0(1)) * (eye - _m_org).norm() * 0.5 * 
	  ((eye - _m_org).unit());
      break;
    default:
      break;
    }
}

/* ************************************************************************* */
void	TrackBall::glInitLists() const
{
  static bool	isListDone(false);

  if (!isListDone || !_s_list)
    {
      _s_list = glGenLists(2);

      glNewList(_s_list, GL_COMPILE);
      glBegin(GL_LINE_STRIP);
      for (double a(0.0); a <= M_2PI; a += 0.15)
	glVertex2d(cos(a), sin(a));
      glVertex2d(1.0, 0.0);
      glEnd();// GL_LINE_STRIP
      glEndList();

      GLUquadric *quad = gluNewQuadric();

      glNewList(_s_list + 1, GL_COMPILE);
      gluSphere(quad, 1.0, 16, 12);
      gluDeleteQuadric(quad);
      glEndList();
      isListDone = true;
    }
}

/* ************************************************************************* */
Point2d	TrackBall::screen2normal(const Point2d &v)
{
  Point2d	ret(0);
 
  ret.x = (v.x / double(_m_v->_m_view(2)) - 0.5) * 2.0 * _m_factw / _m_rad;
  ret.y = (v.y / double(_m_v->_m_view(3)) - 0.5) * 2.0 * _m_facth / _m_rad;
  return ret;
}

/* ************************************************************************* */
Point3d	TrackBall::normal2screen(const Point3d &v)
{
  Point3d	ret;

  ret.x = (v.x * _m_rad / _m_factw + 1.0) * _m_v->_m_view(2) * 0.5;
  ret.y = (v.y * _m_rad / _m_factw + 1.0) * _m_v->_m_view(3) * 0.5;
  ret.z = 0;
  return ret;
}

#ifndef WIN32
/* namespace gl ************************************************************ */
}
#endif
