// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED Tue Jun 22 17:59:29 NZST 2004
// ************************************************************************* //
#ifndef __RIGIDLOG_HH__
#define __RIGIDLOG_HH__

// ************************************************************************* //
#include "gfxTypes.h"

// ************************************************************************* //
/// DESCRIPTION: Represent the logarithm of a rigid affine transformation
///  (combination of rotations and translations). This represetation is 
///  ideal for blending the transformations.
// ************************************************************************* //
class RigidLog
{
  Vector3d	_m_n;
  Vector3d	_m_cxn;
public:
  RigidLog() {}
  RigidLog(const double d): _m_n(d), _m_cxn(d) {}
  RigidLog(const Vector3d &n, const Vector3d &cxn);
  RigidLog	operator+(const RigidLog &rl) const;
  void		operator()(Point3d &p) const;
  Matrix4d	asMatrix() const;
  Matrix4d	exp() const;
  void		exp(Point3d &p) const;
  void		operator*=(const Matrix4d &m);
  void		operator+=(const RigidLog &l);
  RigidLog	operator*(const double d) const
  { return RigidLog(_m_n * d, _m_cxn * d); }
  Matrix4d	pow2() const;
  RigidLog	pow3() const;
};

// ************************************************************************* //
/// DESCRIPTION: Constructor.
// ************************************************************************* //
inline RigidLog::RigidLog(const Vector3d	&n,
			  const Vector3d	&cxn): _m_n(n), _m_cxn(cxn)
{
}

// ************************************************************************* //
/// DESCRIPTION: Addition
// ************************************************************************* //
inline RigidLog	RigidLog::operator+(const RigidLog &rl) const
{
  return RigidLog(_m_n + rl._m_n, _m_cxn + rl._m_cxn);
}

// ************************************************************************* //
/// DESCRIPTION: Applies the exponential of this transformation to the 
///  given point.
///  TODO: check if it's a translation.
// ************************************************************************* //
inline void	RigidLog::operator()(Point3d &p) const
{
  Quaternion		q(_m_n);
  Vector3d		c;

  if ((fabs(_m_n.y) <= fabs(_m_n.x)) && (fabs(_m_n.z) <= fabs(_m_n.x)))
    c = Vector3d(1.0, (_m_n.y - _m_cxn.z) / _m_n.x,
		 (_m_cxn.y + _m_n.z) / _m_n.x);
  else if (fabs(_m_n.z) <= fabs(_m_n.y))
    c = Vector3d((_m_cxn.z + _m_n.x) / _m_n.y, 1.0,
		 (_m_n.z - _m_cxn.x) / _m_n.y);
  else
    c = Vector3d((_m_n.x - _m_cxn.y) / _m_n.z,
		 (_m_cxn.x + _m_n.y) / _m_n.z, 1.0);
  p -= c;
  q.eqExp();
  q(p);
  p += c;
}

// ************************************************************************* //
/// DESCRIPTION: Returns the representation in an equivalent 4x4 matrix form.
// ************************************************************************* //
inline Matrix4d	RigidLog::asMatrix() const
{
  return Matrix4d(0.0, -_m_n.z, _m_n.y, _m_cxn.x,
		  _m_n.z, 0.0, -_m_n.x, _m_cxn.y,
		  -_m_n.y, _m_n.x, 0.0, _m_cxn.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
/// DESCRIPTION: Converts this instance into its exponential form.
/// Three cases: 1. pure translation,
///              2. translation/rotation
///              3. general case:
///  exp(log R) = I + logR + logR^2 (1-cos|n|)/n2 + logR^3(|n|-sin|n|)/(n n2)
// ************************************************************************* //
inline Matrix4d	RigidLog::exp() const
{
  const double	n2(_m_n.norm2());

  if (n2 < 1.0e-16)
    return Translation(_m_cxn).asMatrix();

  const Vector3d	nxcxn(_m_n ^ _m_cxn);

  if (nxcxn.norm2() < 1.0e-64)
    {
      Quaternion	q(_m_n * 0.5);
      Vector3d		c;

      q.eqExp();
      if ((fabs(_m_n.y) <= fabs(_m_n.x)) && (fabs(_m_n.z) <= fabs(_m_n.x)))
	c = Vector3d(1.0, (_m_n.y - _m_cxn.z) / _m_n.x,
		     (_m_cxn.y + _m_n.z) / _m_n.x);
      else if (fabs(_m_n.z) <= fabs(_m_n.y))
	c = Vector3d((_m_cxn.z + _m_n.x) / _m_n.y, 1.0,
		     (_m_n.z - _m_cxn.x) / _m_n.y);
      else
	c = Vector3d((_m_cxn.x + _m_n.y) / _m_n.z,
		     (_m_n.x - _m_cxn.y) / _m_n.z, 1.0);
      return Rotation(q, c).asMatrix();
    }

  const double	n(sqrt(n2)), invn2(1.0 / n2);
  const double	cosn(cos(n)), sinn(sin(n)), k(sinn / n);
  const double	cl2((1.0 - cosn) * invn2), cl3((1.0 - k) * invn2);
  const double	nxy(_m_n.x * _m_n.y * cl2), nxx(_m_n.x * _m_n.x * cl2);
  const double	nyz(_m_n.y * _m_n.z * cl2), nyy(_m_n.y * _m_n.y * cl2);
  const double	nzx(_m_n.z * _m_n.x * cl2), nzz(_m_n.z * _m_n.z * cl2);
  const Vector3d	cn(_m_n * k);
  const Vector3d	ccxn(_m_n * ((_m_cxn * _m_n) * cl3) + _m_cxn * k);

  return Matrix4d(nxx + cosn, nxy - cn.z, nzx + cn.y, nxcxn.x * cl2 + ccxn.x,
		  nxy + cn.z, nyy + cosn, nyz - cn.x, nxcxn.y * cl2 + ccxn.y,
		  nzx - cn.y, nyz + cn.x, nzz + cosn, nxcxn.z * cl2 + ccxn.z,
		  0.0, 0.0, 0.0, 1.0);
}

// ************************************************************************* //
inline void	RigidLog::exp(Point3d &q) const
{
  const double	n2(_m_n.norm2());

  if (n2 < 1.0e-16)
    {
      q += _m_cxn;
      return;
    }

  const Vector3d	nxcxn(_m_n ^ _m_cxn);

  if (nxcxn.norm2() < 1.0e-64)
    {
      Quaternion	qu(_m_n * 0.5);
      Vector3d		c;

      qu.eqExp();
      if ((fabs(_m_n.y) <= fabs(_m_n.x)) && (fabs(_m_n.z) <= fabs(_m_n.x)))
	c = Vector3d(1.0, (_m_n.y - _m_cxn.z) / _m_n.x,
		     (_m_cxn.y + _m_n.z) / _m_n.x);
      else if (fabs(_m_n.z) <= fabs(_m_n.y))
	c = Vector3d((_m_cxn.z + _m_n.x) / _m_n.y, 1.0,
		     (_m_n.z - _m_cxn.x) / _m_n.y);
      else
	c = Vector3d((_m_cxn.x + _m_n.y) / _m_n.z,
		     (_m_n.x - _m_cxn.y) / _m_n.z, 1.0);
      Rotation(qu, c)(q);
      return;
    }

  Vector3d	p(q);
  const double	n(sqrt(n2)), in2(1.0 / n2);
  const double	a(1.0 - cos(n)), b(1.0 - sin(n) / n);

  q += _m_cxn * (1.0 - b) + (_m_n ^ p) * (1.0 + b) + (nxcxn * in2 - p) * a
    + _m_n * (((_m_n * p) * in2) * a + (_m_cxn * _m_n) * in2 * b);
}

// ************************************************************************* //
/// DESCRIPTION: Places the transformation in the coordinate set specified 
///  by matrix m = (x y z c)
///                (0 0 0 1)
/// Constrain: x.x == y.y == z.z and x.y == y.z == z.x == 0
// ************************************************************************* //
inline void	RigidLog::operator*=(const Matrix4d	&m)
{
  const Vector3d	x(m.getx()), y(m.gety()), z(m.getz()), o(m.geto());
  const double		l(x.norm2());
  const double		ox(o * x), oy(o * y), oz(o * z);

  _m_cxn = (_m_cxn.x + (_m_n.z * oy - _m_n.y * oz) / l) * x
    + (_m_cxn.y + (_m_n.x * oz - _m_n.z * ox) / l) * y
    + (_m_cxn.z + (_m_n.y * ox - _m_n.x * oy) / l) * z;
  _m_n = (_m_n.x * (y ^ z) + _m_n.y * (z ^ x) + _m_n.z * (x ^ y)) / l;
}

// ************************************************************************* //
inline void	RigidLog::operator+=(const RigidLog &l)
{
  _m_n += l._m_n;
  _m_cxn += l._m_cxn;
}

// ************************************************************************* //
/// DESCRIPTION: Returns the cube of the matrix.
// ************************************************************************* //
inline Matrix4d	RigidLog::pow2() const
{
  const double	n2(_m_n * _m_n);
  const double	nxy(_m_n.x * _m_n.y);
  const double	nyz(_m_n.y * _m_n.z);
  const double	nzx(_m_n.z * _m_n.x);
  const double	nxx(_m_n.x * _m_n.x);
  const double	nyy(_m_n.y * _m_n.y);
  const double	nzz(_m_n.z * _m_n.z);
  const Vector3d	nxcxn(_m_n ^ _m_cxn);

  return Matrix4d(nxx - n2, nxy, nzx, nxcxn.x,
		  nxy, nyy - n2, nyz, nxcxn.y,
		  nzx, nyz, nzz - n2, nxcxn.z,
		  0.0, 0.0, 0.0, 0.0);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the cube of the matrix.
// ************************************************************************* //
inline RigidLog	RigidLog::pow3() const
{
  const double	n2(-(_m_n * _m_n));

  return RigidLog(_m_n * n2, _m_n * (_m_cxn * _m_n) + _m_cxn * n2);
}

// ************************************************************************* //
#endif//__RIGIDLOG_HH__
