// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFCONJUGATEGRADIENTS_HH__
#define	__SFCONJUGATEGRADIENTS_HH__

#include <time.h>

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Conjugate gradients solver.
// ///////////////////////////////////////////////////////////////////////// //
class CG
{
  SpeedGrid		&_m_xi, &_m_di, &_m_ri, &_m_adi;
  //const UniMatrix	&_m_a;
  const IsoFilter	&_m_a;
public:
  CG(SpeedGrid &xi, SpeedGrid &buf0, SpeedGrid &buf1, SpeedGrid &buf2,
     const IsoFilter &a):
    _m_xi(xi), _m_di(buf0), _m_ri(buf1), _m_adi(buf2), _m_a(a) {}
  double	solve(const unsigned steps = 8,
		      const double epsilon = 0.0) const;
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Returns the residual.
// ///////////////////////////////////////////////////////////////////////// //
inline double	CG::solve(const unsigned	steps,
			  const double		epsilon) const
{
  double	alpha, beta, ri2, rj2;

  _m_di.eqSubMul(_m_xi, _m_a, _m_xi);
  _m_ri.eq(_m_di);
  ri2 = _m_ri.norm2();
  for (unsigned i(0); i < steps && epsilon < ri2; ++i)
    {
      _m_adi.eqMul(_m_a, _m_di);
      alpha = ri2 / _m_di.dot(_m_adi);
      _m_xi.eqAddMul(alpha, _m_di);
      _m_ri.eqSubMul(alpha, _m_adi);
      rj2 = _m_ri.norm2();
      beta = rj2 / ri2;
      _m_di.eqAddMul(_m_ri, beta, _m_di);
      ri2 = rj2;
    }
  return ri2;
}

// ///////////////////////////////////////////////////////////////////////// //
class BackSubstitution
{
  const MultiMatrix	&_m_l;
  DivGrid		&_m_w;//result
  DivGrid		&_m_y;//temporary
  const DivGrid		&_m_z;//input
public:
  BackSubstitution(const MultiMatrix &l, DivGrid &w, 
		   DivGrid &y, const DivGrid &z):
    _m_l(l), _m_w(w), _m_y(y), _m_z(z) {}
  void	solve();
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Solves (_m_l * _m_l^T)* _m_w = _m_z
///  Step 1: Solve in y _m_l * _m_y = _m_z
///  Step 2: Solve in w _m_l^T)* _m_w = _m_y
// ///////////////////////////////////////////////////////////////////////// //
inline void	BackSubstitution::solve()
{
  double	sum;
  int		size(_m_l.size()), trck;
  const int	*trc;

  for (int i(0); i < _m_l.size(); ++i)
    {
      sum = _m_z(i);
      trc = _m_l.getRowColIndex(i);
      for (int k(0); k < 7 && (trck = trc[k]) < i; ++k)
	if (0 <= trck)
	  sum -= _m_l(i, trck) * _m_y(trck);
      _m_y(i) = sum * _m_l.invDiag(i);
    }
  for (int i(size - 1); i >= 0; --i)
    {
      sum = _m_y(i);
      trc = _m_l.getRowColIndex(i);
      for (int k(0); k < 7 && (trck = trc[k]) < size; ++k)
	if (i < trck)
	  sum -= _m_l(trck, i) * _m_w(trck);
      _m_w(i) = sum * _m_l.invDiag(i);
    }
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Cholesky Preconditioned Conjugate Gradients.
// ///////////////////////////////////////////////////////////////////////// //
class PCG
{
  DivGrid		&_m_xi, &_m_di, &_m_ri, &_m_adi, &_m_invmri;
  mutable  IsoFilter	_m_a;
  const MultiMatrix	&_m_l;
public:
  PCG(DivGrid &xi, DivGrid &buf0, DivGrid &buf1, DivGrid &buf2, DivGrid &buf3,
      const IsoFilter &a, const MultiMatrix &l):
    _m_xi(xi), _m_di(buf0), _m_ri(buf1), _m_adi(buf2),
    _m_invmri(buf3), _m_a(a), _m_l(l) {}
  double	solve(const unsigned steps = 8,
		      const double epsilon = 0.0) const;
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Returns the residual.
// ///////////////////////////////////////////////////////////////////////// //
double	PCG::solve(const unsigned	steps,
		   const double		epsilon) const
{
  double		alpha, beta, rmri, rmrj, diadi;
  BackSubstitution	getinvmri(_m_l, _m_invmri, _m_adi, _m_ri);

  _m_ri.eqSubMul(_m_xi, _m_a, _m_xi);
  getinvmri.solve();
  _m_di.eq(_m_invmri);
  rmri = _m_ri.dot(_m_invmri);
  for (unsigned i(0); i < steps && epsilon < rmri; ++i)
    {
      _m_adi.eqMul(_m_a, _m_di);
      diadi = _m_di.dot(_m_adi);
      alpha = rmri / diadi;
      _m_xi.eqAddMul(alpha, _m_di);
      _m_ri.eqSubMul(alpha, _m_adi);
      getinvmri.solve();
      rmrj = _m_ri.dot(_m_invmri);
      beta = rmrj / rmri;
      _m_di.eqAddMul(_m_invmri, beta, _m_di);
      rmri = rmrj;
    }
  return rmri;
}

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFCONJUGATEGRADIENTS_HH__
