// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include "sfSolver.h"
#include "sfSpeedGrid.h"
#include "sfConjugateGradients.h"

// ///////////////////////////////////////////////////////////////////////// //
#ifdef WIN32
#ifdef min
#undef min
#endif
#endif
template <class A, class B>
A	min(A a,
	    B b)
{
  return (a < b)? a: b;
}

// ///////////////////////////////////////////////////////////////////////// //
#ifdef WIN32
#ifdef max
#undef max
#endif
#endif
template <class A, class B>
A	max(A a,
	    B b)
{
  return (b < a)? a: b;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Returns true if a cubic neighborhood of radius 0.5 around p 
///   is contained by the object.
// ///////////////////////////////////////////////////////////////////////// //
#ifdef rad2
#undef rad2
#endif
bool	Object::contains(const Point3d &p) const
{
  const  double		rad2(_m_rad * _m_rad);
  const Vector3d	q(p - _m_p);

  return (q + Vector3d(0.5, 0.5, 0.5)).norm2() <= rad2 &&
    (q + Vector3d(-0.5, 0.5, 0.5)).norm2() <= rad2 &&
    (q + Vector3d(0.5, -0.5, 0.5)).norm2() <= rad2 &&
    (q + Vector3d(-0.5, -0.5, 0.5)).norm2() <= rad2 &&
    (q + Vector3d(0.5, 0.5, -0.5)).norm2() <= rad2 &&
    (q + Vector3d(-0.5, 0.5, -0.5)).norm2() <= rad2 &&
    (q + Vector3d(0.5, -0.5, -0.5)).norm2() <= rad2 &&
    (q + Vector3d(-0.5, -0.5, -0.5)).norm2() <= rad2;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Returns the average normal to the surface in a neighborhood
///   around p.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	Object::normal(const Point3d &p) const
{
  return (p - _m_p).unit();
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Returns true if a cubic neighborhood of radius 0.5 around p 
///   intersects the surface of the object.
// ///////////////////////////////////////////////////////////////////////// //
bool	Object::intersects(const Point3d &p) const
{
  const Point3d	q(_m_p + (p - _m_p).unit() * _m_rad);

  return p.x - 0.5 <= q.x && p.y - 0.5 <= q.y &&
    p.z - 0.5 <= q.z && q.x <= p.x + 0.5 &&
    q.y <= p.y + 0.5 && q.z <= p.z + 0.5;
}

// ///////////////////////////////////////////////////////////////////////// //
bool	Object::touches(const Point3d &p) const
{
  return intersects(p) || contains(p);
}

// ///////////////////////////////////////////////////////////////////////// //
void	Object::glRender() const
{
  glPushMatrix();

  GLUquadric *quad = gluNewQuadric();
  
  _m_p.glTranslate();
  glScaled(_m_rad, _m_rad, _m_rad);
  gluSphere(quad, 1.0, 32, 24);
  gluDeleteQuadric(quad);
  glPopMatrix();
}

// ///////////////////////////////////////////////////////////////////////// //
Solver::Solver(const int	x,
	       const int	y,
	       const int	z,
	       const double	rad,
	       const double	nu,
	       const double	rho):
  _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z),
  _m_speedBuf0(x, y, z),
  _m_speedBuf1(x, y, z),
  _m_speedBuf2(x, y, z),
  _m_bbox(Point3d(-0.5), Point3d(x - 0.5, y - 0.5, z - 0.5)),
  _m_a(6.0, -1.0),
  _m_d(x, y, z),
  _m_object(Point3d((x - 1) * 0.5, (y - 1) * 0.5, (z - 1) * 0.5), rad),
  _m_nu(nu),
  _m_rho(rho)
{
  _m_l = new MultiMatrix(x, y, z), _m_l->cholesky(_m_a);
  _m_speed = new SpeedGrid(x, y, z);
  _m_speedBuf = new SpeedGrid(x, y, z);
  _m_div = new DivGrid(x, y, z);
  _m_pBuf0 = new DivGrid(x, y, z);
  _m_pBuf1 = new DivGrid(x, y, z);
  _m_pBuf2 = new DivGrid(x, y, z);
  _m_pBuf3 = new DivGrid(x, y, z);
}

// ///////////////////////////////////////////////////////////////////////// //
Solver::~Solver()
{
}

// ///////////////////////////////////////////////////////////////////////// //
void	Solver::glRender() const
{
  glEnable(GL_LIGHTING);
  _m_object.glRender();
  glDisable(GL_LIGHTING);
  _m_bbox.glRender();
  _m_speed->glRender();
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Convect.
/// TODO: Stop at interface. Use material density.
// ///////////////////////////////////////////////////////////////////////// //
void	Solver::convect(const double	dt,
			SpeedGrid	*&speed,
			const int	steps)
{
  SpeedGrid	*const swap = speed;

  speed = _m_speedBuf;
  _m_speedBuf = swap;
  speed->convect(*swap, dt, steps);
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Diffuse with CG, with free Neumann boundary condition.
// ///////////////////////////////////////////////////////////////////////// //
void	Solver::diffuse(const double	dt,
			SpeedGrid	&speed)
{
  if (dt == 0.0)
    return;
  
  _m_d.setLine(IsoFilter(1.0 + 6.0 * dt, -dt));
  const IsoFilter	isof(1.0 + 6.0 * dt, -dt);

  CG	cjDiffuse(speed, _m_speedBuf0, _m_speedBuf1, _m_speedBuf2, isof);
  
  std::cout << "Diffuse residual = " << cjDiffuse.solve(128, 1.0e-6)
	    << std::endl;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Diffuse with CG, with free Neumann boundary condition.
// ///////////////////////////////////////////////////////////////////////// //
void	Solver::blur(SpeedGrid		*&speed)
{
  SpeedGrid	*const swap = speed;

  speed = _m_speedBuf;
  _m_speedBuf = swap;
  speed->blur(*swap);
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Projects with a CG.
///  0. 
///  1. compute divergence as a pressure. None on container... 
///  2. substract gradient from field, but not container (because = 0).
// ///////////////////////////////////////////////////////////////////////// //
void	Solver::project(SpeedGrid	&speed)
{
  speed.keepInside();
  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	(*_m_div)(x, y, z) = 
	  (speed.cut(x - 1, y, z).x - speed.cut(x + 1, y, z).x + 
	   speed.cut(x, y - 1, z).y - speed.cut(x, y + 1, z).y + 
	   speed.cut(x, y, z - 1).z - speed.cut(x, y, z + 1).z) * 0.5 * _m_rho;

  PCG	cjProject(*_m_div, *_m_pBuf0, *_m_pBuf1, *_m_pBuf2, *_m_pBuf3,
		  _m_a, *_m_l);
  
  std::cout << "Project residual = " << cjProject.solve(128, 1.0e-9)
	    << std::endl;

  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	speed(x, y, z) -= 
	  Vector3d(_m_div->checkBoundary(x + 1, y, z) -
		   _m_div->checkBoundary(x - 1, y, z),
		   _m_div->checkBoundary(x, y + 1, z) -
		   _m_div->checkBoundary(x, y - 1, z),
		   _m_div->checkBoundary(x, y, z + 1) -
		   _m_div->checkBoundary(x, y, z - 1)) * 0.5 / _m_rho;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Marches through time with dt, in seconds.
/// STEPS:
///  1. Create ghost speeds.
///    1.1. Split speed with phi in 2 ghost speeds.
///    1.2. Extrapolate average ghost speeds by following phi gradient.
///    1.3. convect phi with speed.
///  2. Solve each ghost speeds.
///    3.1. add force (except on boundary)
///    3.2. convect
///    3.3. diffuse
///    3.4. project
///  3. Merge ghost speeds and create new speed.
// ///////////////////////////////////////////////////////////////////////// //
void	Solver::solve(const double dt)
{
  convect(dt, _m_speed, 8);
  diffuse(dt * _m_nu, *_m_speed);
  project(*_m_speed);
}

// ///////////////////////////////////////////////////////////////////////// //
void	Solver::maskObject()
{
  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	if (!_m_object.touches(Point3d(x, y, z)))
	  (*_m_speed)(x, y, z) = Vector3d(0.0);
}

// ///////////////////////////////////////////////////////////////////////// //
void	Solver::operator()(Point3d	&p,
			   const double	dt,
			   const int	steps) const
{
  const double	inc(dt / steps);

  for (int i(0); i < steps; ++i)
    p += _m_speed->bspline(p) * inc;
}

// ///////////////////////////////////////////////////////////////////////// //
Vector3d	&Solver::operator()(const int x,
				    const int y,
				    const int z)
{
  return (*_m_speed)(x, y, z);
}

// ///////////////////////////////////////////////////////////////////////// //
Vector3d	&Solver::operator()(const int i)
{
  return (*_m_speed)(i);
}

// ///////////////////////////////////////////////////////////////////////// //
void	Solver::reset()
{
  for (int i(0); i < _m_size; ++i)
    (*_m_speed)(i) = Vector3d(0.0);
}

// ///////////////////////////////////////////////////////////////////////// //
Vector3d	Solver::speed(const Point3d &p) const
{
  return _m_speed->bspline(p);
}
