// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFSPARSEMATRIX_HH___
#define	__SFSPARSEMATRIX_HH___

// ///////////////////////////////////////////////////////////////////////// //
#include <vector>
#include "sfFilter.h"
#include "sfIsoFilter.h"

// ///////////////////////////////////////////////////////////////////////// //
class SparseMatrix
{
public:
  virtual int	size() const = 0;
  virtual int	density() const = 0;
  virtual double	&operator()(const int i, const int j) = 0;
  virtual double	operator()(const int i, const int j) const = 0;
  virtual void		cholesky(const IsoFilter &f) = 0;
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Sparse matrix.
// ///////////////////////////////////////////////////////////////////////// //
class UniMatrix: public SparseMatrix
{
  const int	_m_x, _m_y, _m_z, _m_size;
  IsoFilter	_m_lines;
  static int	*_s_rowCol;
public:
  UniMatrix(const IsoFilter &f, const int x, const int y, const int z):
    _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z), _m_lines(f)
  { precomputeRowCol(_m_x, _m_y, _m_z, _m_size); }
  UniMatrix(const int x, const int y, const int z):
    _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z)
  { precomputeRowCol(_m_x, _m_y, _m_z, _m_size); }
  double	&operator()(const int i, const int j);
  double	operator()(const int i, const int j) const;
  friend std::ostream	&operator<<(std::ostream &os, const UniMatrix &uf);
  int	size() const { return _m_size; }
  int	density() const { return 7; }
  static inline const int	*getRowColIndex(const int i) 
  { return _s_rowCol + 7 * i; }
  static void	precomputeRowCol(const int size, const int x, const int y, 
				 const int z);
  void	cholesky(const IsoFilter &) {}
  void	setLine(const IsoFilter &lines) { _m_lines = lines; }
};

// ///////////////////////////////////////////////////////////////////////// //
inline std::ostream	&operator<<(std::ostream	&os,
				    const UniMatrix	&uf)
{
  for (int i(0); i < uf._m_size; ++i)
    {
      for (int j(0); j < uf._m_size; ++j)
	os << uf(i, j) << " ";
      os << "\n";
    }
  return os;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	&UniMatrix::operator()(const int	i,
				       const int	j)
{
  if (i == j)
    return _m_lines(1, 1, 1);

  const int	*itab;

  itab = getRowColIndex(i);
  if (j == itab[0])
    return _m_lines._m_s;
  else if (j == itab[1])
    return _m_lines._m_s;
  else if (j == itab[2])
    return _m_lines._m_s;
  else if (j == itab[4])
    return _m_lines._m_s;
  else if (j == itab[5])
    return _m_lines._m_s;
  else if (j == itab[6])
    return _m_lines._m_s;

  static double	zero;

  zero = 0.0;
  return zero;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	UniMatrix::operator()(const int	i,
				      const int	j) const
{
  if (i == j)
    return _m_lines(1, 1, 1);

  const int	*itab;

  itab = getRowColIndex(i);
  if (j == itab[0])
    return _m_lines._m_s;
  else if (j == itab[1])
    return _m_lines._m_s;
  else if (j == itab[2])
    return _m_lines._m_s;
  else if (j == itab[4])
    return _m_lines._m_s;
  else if (j == itab[5])
    return _m_lines._m_s;
  else if (j == itab[6])
    return _m_lines._m_s;
  return 0.0;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Sparse matrix in the form of a filter.
// ///////////////////////////////////////////////////////////////////////// //
class MultiMatrix: public SparseMatrix
{
public:
  const int	_m_x, _m_y, _m_z, _m_size;
  Filter	*const _m_lines;
  double	*_m_invDiag; //inversed diagonal.
  static int	*_s_rowCol;
public:
  MultiMatrix(const int x, const int y, const int z):
    _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z),
    _m_lines(new Filter[_m_size]), _m_invDiag(new double[_m_size])
  { precomputeRowCol(_m_x, _m_y, _m_z, _m_size); }
  virtual ~MultiMatrix() { delete[] _m_lines; delete[] _m_invDiag; }
  inline double	&operator()(const int i, const int j);
  inline double	operator()(const int i, const int j) const;
  Filter	&operator()(const int i) { return _m_lines[i]; }
  void		reset();
  void		cholesky(const IsoFilter &f);
  void		cholesky(const MultiMatrix &f);
  friend std::ostream	&operator<<(std::ostream &os, const MultiMatrix &uf);
  inline int		size() const { return _m_size; }
  int		density() const { return 7; }
  static inline const int	*getRowColIndex(const int i) 
  { return _s_rowCol + 7 * i; }
  static void	precomputeRowCol(const int size, const int x, const int y, 
				 const int z);
  void		precomputeInvDiag();
  inline double	invDiag(const int i) const { return _m_invDiag[i]; }
};

// ///////////////////////////////////////////////////////////////////////// //
inline void	MultiMatrix::precomputeInvDiag()
{
  for (int i(0); i < _m_size; ++i)
    _m_invDiag[i] = 1.0 / operator()(i, i);
}

// ///////////////////////////////////////////////////////////////////////// //
inline void	MultiMatrix::precomputeRowCol(const int	,
					      const int	y,
					      const int	z,
					      const int	size)
{
  static bool	done(false);

  if (done)
    return;
  _s_rowCol = new int[7 * size];
  for (int i(0); i < size; ++i)
    {
      _s_rowCol[i * 7 + 3] = i;
      _s_rowCol[i * 7 + 0] = i - y * z;
      _s_rowCol[i * 7 + 1] = i - z;
      _s_rowCol[i * 7 + 2] = i - 1;
      _s_rowCol[i * 7 + 4] = i + 1;
      _s_rowCol[i * 7 + 5] = i + z;
      _s_rowCol[i * 7 + 6] = i + y * z;
    }
  done = true;
}

// ///////////////////////////////////////////////////////////////////////// //
inline std::ostream	&operator<<(std::ostream	&os,
				    const MultiMatrix	&mf)
{
  for (int i(0); i < mf._m_size; ++i)
    {
      for (int j(0); j < mf._m_size; ++j)
	os << mf(i, j) << " ";
      os << "\n";
    }
  return os;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	&MultiMatrix::operator()(const int i,
					 const int j)
{
  if (i == j)
    return _m_lines[i](1, 1, 1);

  const int	*itab;

  itab = getRowColIndex(i);
  if (j == itab[0])
    return _m_lines[i]._m_s[0];
  else if (j == itab[1])
    return _m_lines[i]._m_s[2];
  else if (j == itab[2])
    return _m_lines[i]._m_s[4];
  else if (j == itab[4])
    return _m_lines[i]._m_s[5];
  else if (j == itab[5])
    return _m_lines[i]._m_s[3];
  else if (j == itab[6])
    return _m_lines[i]._m_s[1];

  static double	zero;

  zero = 0.0;
  return zero;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	MultiMatrix::operator()(const int i,
					const int j) const
{
  if (i == j)
    return _m_lines[i](1, 1, 1);

  const int	*itab;

  itab = getRowColIndex(i);
  if (j == itab[0])
    return _m_lines[i]._m_s[0];
  else if (j == itab[1])
    return _m_lines[i]._m_s[2];
  else if (j == itab[2])
    return _m_lines[i]._m_s[4];
  else if (j == itab[4])
    return _m_lines[i]._m_s[5];
  else if (j == itab[5])
    return _m_lines[i]._m_s[3];
  else if (j == itab[6])
    return _m_lines[i]._m_s[1];
  return 0.0;
}

// ///////////////////////////////////////////////////////////////////////// //
inline void	MultiMatrix::reset()
{
  for (int i(0); i < _m_size; ++i)
    _m_lines[i].reset();
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Incomplete Cholesky. The matrix is given by the filter f.
// ///////////////////////////////////////////////////////////////////////// //
inline void	MultiMatrix::cholesky(const IsoFilter &f)
{
  double		sum;
  const int		*trci, *trcj;
  const UniMatrix	uf(f, _m_x, _m_y, _m_z);

  reset();
  for (int i(0); i < _m_size; ++i)
    {
      trci = getRowColIndex(i);
      for (int ji(0); ji < 7; ++ji)
	if (i <= trci[ji] && trci[ji] < _m_size)
	  {
	    sum = uf(i, trci[ji]);
	    trcj = getRowColIndex(trci[ji]);
	    for (int kj(0); kj < 7; ++kj)
	      for (int ki(0); ki < 7; ++ki)
		if (0 <= trcj[kj] && trcj[kj] < i && 
		    0 <= trci[ki] && trci[ki] < i &&
		    trci[ki] == trcj[kj])
		  sum -= uf(i, trci[ki]) * uf(trci[ji], trcj[kj]);
	    if (i == trci[ji])
	      operator()(i, i) = sqrt(sum);
	    else
	      operator()(trci[ji], i) = sum / operator()(i, i);
	  }
    }
  precomputeInvDiag();
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Incomplete Cholesky of sparse matrix mf.
// ///////////////////////////////////////////////////////////////////////// //
inline void	MultiMatrix::cholesky(const MultiMatrix &mf)
{
  double	sum;
  const int	*trci, *trcj;

  reset();
  for (int i(0); i < _m_size; ++i)
    {
      trci = getRowColIndex(i);
      for (int ji(0); ji < 7; ++ji)
	if (i <= trci[ji] && trci[ji] < _m_size)
	  {
	    sum = mf(i, trci[ji]);
	    trcj = getRowColIndex(trci[ji]);
	    for (int kj(0); kj < 7; ++kj)
	      for (int ki(0); ki < 7; ++ki)
		if (0 <= trcj[kj] && trcj[kj] < i && 
		    0 <= trci[ki] && trci[ki] < i &&
		    trci[ki] == trcj[kj])
		  sum -= mf(i, trci[ki]) * mf(trci[ji], trcj[kj]);
	    if (i == trci[ji])
	      {
		std::cout << "> " << sum << std::endl;
		operator()(i, i) = sqrt(sum);
	      }
	    else
	      operator()(trci[ji], i) = sum / operator()(i, i);
	  }
    }
  precomputeInvDiag();
}

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFSPARSEMATRIX_HH__
