// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#include "sfSpeedGrid.h"
#include "sfIsoFilter.h"
#include "sfFilter.h"
#include "sfSparseMatrix.h"

// ///////////////////////////////////////////////////////////////////////// //
int		*DivGrid::_s_nghbd(0);
SpeedGrid	*SpeedGrid::_s_dk, *SpeedGrid::_s_deltak;
int		*SpeedGrid::_s_nghbd(0);

// ///////////////////////////////////////////////////////////////////////// //
DivGrid::DivGrid(const int x,
		 const int y,
		 const int z):
  _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z), _m_grid(new double[_m_size])
{
  precomputeNghbd(x, y, z, _m_size);
}

// ///////////////////////////////////////////////////////////////////////// //
void	DivGrid::precomputeNghbd(const int mx,
				 const int my,
				 const int mz,
				 const int size)
{
  static bool	done(false);
  const int	yz(my * mz);

  if (done)
    return;
  _s_nghbd = new int[6 * size];
  for (int i(0); i < size; ++i)
    {
      const int tmp(i / mz);
      const int z(i % mz);
      const int y(tmp % my);

      _s_nghbd[6 * i + 0] = ((tmp < my)? i: (i - yz));
      _s_nghbd[6 * i + 1] = ((mx - 1 == tmp / my)? i: (i + yz));
      _s_nghbd[6 * i + 2] = ((y == 0)? i: (i - mz));
      _s_nghbd[6 * i + 3] = ((my - 1 <= y)? i: (i + mz));
      _s_nghbd[6 * i + 4] = ((z == 0)? i: i - 1);
      _s_nghbd[6 * i + 5] = ((mz - 1 == z)? i: i + 1);
    }
  done = true;
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::precomputeNghbd(const int mx,
				   const int my,
				   const int mz,
				   const int size)
{
  static bool	done(false);
  const int	yz(my * mz);

  if (done)
    return;
  _s_nghbd = new int[6 * size];
  for (int i(0); i < size; ++i)
    {
      const int tmp(i / mz);
      const int z(i % mz);
      const int y(tmp % my);

      _s_nghbd[6 * i + 0] = ((tmp < my)? i: (i - yz));
      _s_nghbd[6 * i + 1] = ((mx - 1 == tmp / my)? i: (i + yz));
      _s_nghbd[6 * i + 2] = ((y == 0)? i: (i - mz));
      _s_nghbd[6 * i + 3] = ((my - 1 <= y)? i: (i + mz));
      _s_nghbd[6 * i + 4] = ((z == 0)? i: i - 1);
      _s_nghbd[6 * i + 5] = ((mz - 1 == z)? i: i + 1);
    }
  done = true;
}

// ///////////////////////////////////////////////////////////////////////// //
void	DivGrid::eq(const DivGrid &sg)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] = sg._m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
double	&DivGrid::operator()(const int x,
			     const int y,
			     const int z)
{
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
double	DivGrid::operator()(const int x,
			    const int y,
			    const int z) const
{
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
double	DivGrid::checkBoundary(const int x,
			       const int y,
			       const int z) const
{
  if (x < 0)
    return operator()(0, y, z);
  if (y < 0)
    return operator()(x, 0, z);
  if (z < 0)
    return operator()(x, y, 0);
  if (_m_x <= x)
    return operator()(_m_x - 1, y, z);
  if (_m_y <= y)
    return operator()(x, _m_y - 1, z);
  if (_m_z <= z)
    return operator()(x, y, _m_z - 1);
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
double	SpeedGrid::norm2() const
{
  double	ret(0.0);

  for (int i(0); i < _m_size; ++i)
    ret += _m_grid[i] * _m_grid[i];
  return ret;
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eq(const SpeedGrid &sg)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] = sg._m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Renders per cube the average of the speed for the pairs 
///  of oposit squares.
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::glRender() const
{
  Vector3d	v;
  double	d;

  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	{
	  v = operator()(x, y, z);
	  if (fabs(d = v.norm()) > 1.0e-6)
	    {
	      //v /= d;
	      glBegin(GL_LINES);
	      glColor3d(d, 0.0, 0.0);
	      glVertex3d(x, y, z);
	      glColor3d(d, 0.33, 0.33);
	      (Point3d(x, y, z) + v).glVertex();
	      glEnd();//GL_LINES
	    }
	}
}

// ///////////////////////////////////////////////////////////////////////// //
Vector3d	SpeedGrid::trilin(const Point3d &p) const
{
  const int	i = int(floor(p.x));
  const int	j = int(floor(p.y));
  const int	k = int(floor(p.z));
  const double		u(p.x - i), v(p.y - j), w(p.z - k);
  const Vector3d	vi00(extrapolate(i, j, k) * (1.0 - u) + 
			     extrapolate(i + 1, j, k) * u);
  const Vector3d	vi10(extrapolate(i, j + 1, k) * (1.0 - u) + 
			     extrapolate(i + 1, j + 1, k) * u);
  const Vector3d	vi11(extrapolate(i, j + 1, k + 1) * (1.0 - u) + 
			     extrapolate(i + 1, j + 1, k + 1) * u);
  const Vector3d	vi01(extrapolate(i, j, k + 1) * (1.0 - u) + 
			     extrapolate(i + 1, j, k + 1) * u);

  return (vi00 * (1.0 - v) + vi10 * v) * (1.0 - w) +
    (vi01 * (1.0 - v) + vi11 * v) * w;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Stable interpolant.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	SpeedGrid::monotonic(const Point3d &p) const
{
  const int	tx = int(floor(p.x));
  const int	ty = int(floor(p.y));
  const int	tz = int(floor(p.z));
  const int	x0(tx - 1), x2(tx + 1), x3(tx + 2);
  const double	ux(p.x - tx), uy(p.y - ty), uz(p.z - tz);
  Vector3d	tabx[4][4], taby[4];

  if (0 < tx && 0 < ty && 0 < tz &&
      tx < _m_x - 2 && ty < _m_y - 2 && tz < _m_z - 2)
    for (int y(ty - 1); y < ty + 3; ++y)
      for (int z(tz - 1); z < tz + 3; ++z)
	{
	  const Vector3d fk[4] = { operator()(x0, y, z), operator()(tx, y, z),
				   operator()(x2, y, z), operator()(x3, y, z)};
	  tabx[z + 1 - tz][y + 1 - ty] = coeff(fk, ux);
	}
  else
    for (int y(ty - 1); y < ty + 3; ++y)
      for (int z(tz - 1); z < tz + 3; ++z)
	{
	  const Vector3d fk[4] = { cut(x0, y, z), cut(tx, y, z),
				   cut(x2, y, z), cut(x3, y, z) };
	  tabx[z + 1 - tz][y + 1 - ty] = coeff(fk, ux);
	}
  for (int z(0); z < 4; ++z)
    taby[z] = coeff(tabx[z], uy);
  return coeff(taby, uz);
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: interpolates the speed.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	SpeedGrid::bspline(const Point3d &p) const
{
  const int	tx = int(floor(p.x));
  const int	ty = int(floor(p.y));
  const int	tz = int(floor(p.z));
  const int	x0(tx - 1), x2(tx + 1), x3(tx + 2);
  const double	ux(p.x - tx), uy(p.y - ty), uz(p.z - tz);
  Vector3d	tabx[4][4], taby[4];
  double	bsp0, bsp1, bsp2, bsp3, x, mx;

  x = ux;
  mx = 1.0 - x;
  bsp0 = mx * mx * mx / 6.0;
  bsp1 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * mx;
  bsp2 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * x;
  bsp3 =  x * x * x / 6.0;
  if (0 < tx && 0 < ty && 0 < tz && 
      tx < _m_x - 2 && ty < _m_y - 2 && tz < _m_z - 2)
    for (int y(ty - 1); y < ty + 3; ++y)
      for (int z(tz - 1); z < tz + 3; ++z)
	{
	  const Vector3d fk[4] = { operator()(x0, y, z), operator()(tx, y, z),
				   operator()(x2, y, z), operator()(x3, y, z)};

	  tabx[z + 1 - tz][y + 1 - ty] = bsp0 * fk[0] + bsp1 * fk[1] + 
	    bsp2 * fk[2] + bsp3 * fk[3];
	}
  else
    for (int y(ty - 1); y < ty + 3; ++y)
      for (int z(tz - 1); z < tz + 3; ++z)
	{
	  const Vector3d fk[4] = { cut(x0, y, z), cut(tx, y, z),
				   cut(x2, y, z), cut(x3, y, z) };

	  tabx[z + 1 - tz][y + 1 - ty] = bsp0 * fk[0] + bsp1 * fk[1] + 
	    bsp2 * fk[2] + bsp3 * fk[3];
	}
  x = uy;
  mx = 1.0 - x;
  bsp0 = mx * mx * mx / 6.0;
  bsp1 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * mx;
  bsp2 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * x;
  bsp3 =  x * x * x / 6.0;
  for (int z(0); z < 4; ++z)
    taby[z] = bsp0 * tabx[z][0] + bsp1 * tabx[z][1] + 
      bsp2 * tabx[z][2] + bsp3 * tabx[z][3];
  x = uz;
  mx = 1.0 - x;
  bsp0 = mx * mx * mx / 6.0;
  bsp1 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * mx;
  bsp2 = 1.0 / 6.0 + 0.5 * (1 + x * mx) * x;
  bsp3 =  x * x * x / 6.0;
  return bsp0 * taby[0] + bsp1 * taby[1] + bsp2 * taby[2] + bsp3 * taby[3];
}

// ///////////////////////////////////////////////////////////////////////// //
const Vector3d	&SpeedGrid::extrapolate(const int x,
					const int y,
					const int z) const
{
  if (x < 0)
    return operator()(0, y, z);
  if (y < 0)
    return operator()(x, 0, z);
  if (z < 0)
    return operator()(x, y, 0);
  if (_m_x <= x)
    return operator()(_m_x - 1, y, z);
  if (_m_y <= y)
    return operator()(x, _m_y - 1, z);
  if (_m_z <= z)
    return operator()(x, y, _m_z - 1);
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
const Vector3d	&SpeedGrid::cut(const int x,
				const int y,
				const int z) const
{
  static const Vector3d	zero(0.0);

  if (x < 0 || z < 0 || y < 0 || _m_x <= x || _m_y <= y || _m_z <= z)
    return zero;
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	&SpeedGrid::operator()(const int x,
				       const int y,
				       const int z)
{
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
const Vector3d	&SpeedGrid::operator()(const int x,
				       const int y,
				       const int z) const
{
  return _m_grid[(x * _m_y + y) * _m_z + z];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	&SpeedGrid::operator()(const int i)
{
  return _m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Speed outside is 0, and is also 0 on surface.
// ///////////////////////////////////////////////////////////////////////// //
const Vector3d	&SpeedGrid::operator()(const int i) const
{
  return _m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::blur(SpeedGrid &b)
{
  const int	*trcj;

  for (int i(0); i < _m_size; ++i)
    {
      _m_grid[i] = b(i);
      trcj = DivGrid::getNghbd(i);
      for (int j(0); j < 6; ++j)
	_m_grid[i] += b(trcj[j]);
      _m_grid[i] /= 7;
    }
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Convects the speed sg. Result is put in this.
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::convect(SpeedGrid	&sg,
			   const double	dt,
			   const int	steps)
{
  const double	inc(dt / steps);
  const double	invInc(steps / dt);
  Point3d	p, q;

  for (int i(0); i < _m_size; ++i)
    sg(i) *= inc;
  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	{
	  p = Point3d(x, y, z);
	  for (int i(0); i < steps; ++i)
	    {
	      p -= sg.monotonic(p);
	      if (p.x < 0.0 || p.y < 0.0 || p.z < 0.0 ||
		  _m_x - 1.0 < p.x || _m_y - 1.0 < p.y || _m_z - 1.0 < p.z)
		break;
	    }
	  operator()(x, y, z) = sg.monotonic(p) * invInc;
	}
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this = b - A * w, where * is a convolution with the inside
///  (ignore this and b on the outside).
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqSubMul(const SpeedGrid	&b,
			    const Filter	&a,
			    const SpeedGrid	&w)
{
  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	operator()(x, y, z) = b.operator()(x, y, z) - a.convolve(w, x, y, z);
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this -= d * w
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqSubMul(const double	d,
			    const SpeedGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] -= d * w._m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this = b - A * w.
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqSubMul(const SpeedGrid	&b,
			    const MultiMatrix	&a,
			    const SpeedGrid	&w)
{
  const int	*trcj;

  for (int i(0); i < _m_size; ++i)
    {
      _m_grid[i] = b(i);
      trcj = a.getRowColIndex(i);
      for (int j(0); j < 7; ++j)
	if (0 <= trcj[j] && trcj[j] < _m_size)
	  _m_grid[i] -= a(i, trcj[j]) * w._m_grid[trcj[j]];
    }
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this = b - A * w.
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqSubMul(const SpeedGrid	&b,
			    const UniMatrix	&a,
			    const SpeedGrid	&w)
{
  const int	*trcj;

  for (int i(0); i < _m_size; ++i)
    {
      _m_grid[i] = b(i);
      trcj = a.getRowColIndex(i);
      for (int j(0); j < 7; ++j)
	if (0 <= trcj[j] && trcj[j] < _m_size)
	  _m_grid[i] -= a(i, trcj[j]) * w._m_grid[trcj[j]];
    }
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this = b - A * w.
// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqSubMul(const SpeedGrid	&b,
			    const IsoFilter	&a,
			    const SpeedGrid	&w)
{
  const double	s(a._m_s), c(a._m_c);
  const int	*nghbd;

  for (int i(0); i < _m_size; ++i)
    {
      /*
      nghbd = getNghbd(i);
      _m_grid[i] = b._m_grid[i] - w._m_grid[i] * c - 
	(w._m_grid[nghbd[0]] + w._m_grid[nghbd[1]] + w._m_grid[nghbd[2]] +
	 w._m_grid[nghbd[3]] + w._m_grid[nghbd[4]] + w._m_grid[nghbd[5]]) * s;
      */
      nghbd = getNghbd(i);
      _m_grid[i] = b._m_grid[i] - w._m_grid[i] * c;
      for (int j(0); j < 6; ++j)
	if (nghbd[j] != i)
	  _m_grid[i] -= w._m_grid[nghbd[j]] * s;
    }
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Convolution, without checking boundaries.
// ///////////////////////////////////////////////////////////////////////// //
Vector3d	Filter::convolve(const SpeedGrid	&w,
				 const int		x,
				 const int		y, 
				 const int		z) const
{
  return w(x, y, z) * _m_c
    + w(x - 1, y, z) * _m_s[0] + w(x + 1, y, z) * _m_s[1]
    + w(x, y - 1, z) * _m_s[2] + w(x, y + 1, z) * _m_s[3]
    + w(x, y, z - 1) * _m_s[4] + w(x, y, z + 1) * _m_s[5];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Convolution, without checking boundaries.
// ///////////////////////////////////////////////////////////////////////// //
double	Filter::convolve(const DivGrid	&w,
			 const int	x,
			 const int	y,
			 const int	z) const
{
  return w(x, y, z) * _m_c
    + w(x - 1, y, z) * _m_s[0] + w(x + 1, y, z) * _m_s[1]
    + w(x, y - 1, z) * _m_s[2] + w(x, y + 1, z) * _m_s[3]
    + w(x, y, z - 1) * _m_s[4] + w(x, y, z + 1) * _m_s[5];
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqMul(const Filter		&a,
			 const SpeedGrid	&w)
{
  for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
      for (int z(0); z < _m_z; ++z)
	operator()(x, y, z) = a.convolve(w, x, y, z);
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqMul(const MultiMatrix	&a,
			 const SpeedGrid	&w)
{
  const int	*trcj;

  for (int i(0); i < _m_size; ++i)
    {
      operator()(i) = 0.0;
      trcj = a.getRowColIndex(i);
      for (int j(0); j < a.density(); ++j)
	if (0 <= trcj[j] && trcj[j] < _m_size)
	  operator()(i) += a(i, trcj[j]) * w._m_grid[trcj[j]];
    }
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqMul(const UniMatrix	&a,
			 const SpeedGrid	&w)
{
  const int	*trcj;

  for (int i(0); i < _m_size; ++i)
    {
      operator()(i) = 0.0;
      trcj = a.getRowColIndex(i);
      for (int j(0); j < a.density(); ++j)
	if (0 <= trcj[j] && trcj[j] < _m_size)
	  operator()(i) += a(i, trcj[j]) * w._m_grid[trcj[j]];
    }
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqMul(const IsoFilter	&a,
			 const SpeedGrid	&w)
{
  const double	s(a._m_s), c(a._m_c);
  const int	*nghbd;

  for (int i(0); i < _m_size; ++i)
    {
      /*
      nghbd = getNghbd(i);
      _m_grid[i] = w._m_grid[i] * c + 
	(w._m_grid[nghbd[0]] + w._m_grid[nghbd[1]] + w._m_grid[nghbd[2]] +
	 w._m_grid[nghbd[3]] + w._m_grid[nghbd[4]] + w._m_grid[nghbd[5]]) * s;
      */
      nghbd = getNghbd(i);
      _m_grid[i] = w._m_grid[i] * c;
      for (int j(0); j < 6; ++j)
	if (nghbd[j] != i)
	  _m_grid[i] += w._m_grid[nghbd[j]] * s;
    }
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqAddMul(const SpeedGrid	&b,
			    const double	d,
			    const SpeedGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] = b._m_grid[i] + w._m_grid[i] * d;
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::eqAddMul(const double	d,
			    const SpeedGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] += w._m_grid[i] * d;
}

// ///////////////////////////////////////////////////////////////////////// //
double	SpeedGrid::dot(const SpeedGrid &w) const
{
  double	ret(0.0);

  for (int i(0); i < _m_size; ++i)
    ret += _m_grid[i] * w._m_grid[i];
  return ret;
}

// ///////////////////////////////////////////////////////////////////////// //
void	SpeedGrid::keepInside()
{
  for (int y(0); y < _m_y; ++y)
    for (int z(0); z < _m_z; ++z)
      operator()(0, y, z).x = operator()(_m_x - 1, y, z).x = 0.0;
  for (int x(0); x < _m_x; ++x)
    {
      for (int z(0); z < _m_z; ++z)
	operator()(x, 0, z).y = operator()(x, _m_y - 1, z).y = 0.0;
      for (int y(0); y < _m_y; ++y)
	operator()(x, y, 0).z = operator()(x, y, _m_z - 1).z = 0.0;
    }
  /*
    for (int y(0); y < _m_y; ++y)
    for (int z(0); z < _m_z; ++z)
    {
    if (operator()(0, y, z).x < 0.0)
    operator()(0, y, z).x = 0.0;
    if (0.0 < operator()(_m_x - 1, y, z).x)
    operator()(_m_x - 1, y, z).x = 0.0;
    }
    for (int z(0); z < _m_z; ++z)
    for (int x(0); x < _m_x; ++x)
    {
    if (operator()(x, 0, z).y < 0.0)
    operator()(x, 0, z).y = 0.0;
    if (0.0 < operator()(x, _m_y - 1, z).y)
    operator()(x, _m_y - 1, z).y = 0.0;
    }
    for (int x(0); x < _m_x; ++x)
    for (int y(0); y < _m_y; ++y)
    {
    if (operator()(x, y, 0).z < 0.0)
    operator()(x, y, 0).z = 0.0;
    if (0.0 < operator()(x, y, _m_z - 1).z)
    operator()(x, y, _m_z - 1).z = 0.0;
    }
  */
}
