// ///////////////////////////////////////////////////////////////////////// //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Feb 10 22:39:20 NZDT 2004
// ///////////////////////////////////////////////////////////////////////// //
#ifndef	__SFSPEEDGRID_HH__
#define	__SFSPEEDGRID_HH__

// ///////////////////////////////////////////////////////////////////////// //
#include "gfxTypes.h"
#include "sfIsoFilter.h"

// ///////////////////////////////////////////////////////////////////////// //
class IsoFilter;
class Filter;
class MultiMatrix;
class UniMatrix;

// ///////////////////////////////////////////////////////////////////////// //
class DivGrid
{
protected:
  const int	_m_x, _m_y, _m_z, _m_size;
  double	*const _m_grid;
  static int	*_s_nghbd;
public:
  DivGrid(const int x, const int y, const int z);
  virtual ~DivGrid() { delete[] _m_grid; }
  inline double	&operator()(const int i) { return _m_grid[i]; }
  inline double	operator()(const int i) const { return _m_grid[i]; }
  double	&operator()(const int x, const int y, const int z);
  double	operator()(const int x, const int y, const int z) const;
  double	checkBoundary(const int x, const int y, const int z) const;
  void		eq(const DivGrid &sg);
  inline void		eqSubMul(const double d, const DivGrid &w);
  inline void	eqSubMul(const DivGrid &b, const IsoFilter &a,
			 const DivGrid &w);
  inline void	eqAddMul(const DivGrid &b, const double d, const DivGrid &w);
  inline void	eqAddMul(const double d, const DivGrid &w);
  inline void	eqMul(const IsoFilter &a, const DivGrid &w);
  inline double	norm2() const;
  inline double	dot(const DivGrid &w) const;
  static void	precomputeNghbd(const int x, const int y, const int z,
				const int size);
  static const int	*getNghbd(const int i) { return _s_nghbd + 6 * i; }
};

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Grid with speed defined at the center of faces.
///  On the last faces in x, tbe y and z values are just ignored
///  (same thing for y and for z).
///  There is no speed outside.
// ///////////////////////////////////////////////////////////////////////// //
class SpeedGrid
{
protected:
  const int	_m_x, _m_y, _m_z, _m_size;
  Vector3d	*const _m_grid;
  static SpeedGrid	*_s_dk, *_s_deltak;
  static int	*_s_nghbd;
public:
  SpeedGrid(const int x, const int y, const int z):
    _m_x(x), _m_y(y), _m_z(z), _m_size(x * y * z), 
    _m_grid(new Vector3d[_m_size]) { precomputeNghbd(x, y, z, _m_size); }
  virtual ~SpeedGrid() { delete[] _m_grid; }
  Vector3d	bspline(const Point3d &p) const;
  Vector3d	monotonic(const Point3d &p) const;
  Vector3d	trilin(const Point3d &p) const;
  Vector3d	&operator()(const int i);
  const Vector3d&operator()(const int i) const;
  Vector3d	&operator()(const int x, const int y, const int z);
  const Vector3d&operator()(const int x, const int y, const int z) const;
  const Vector3d&extrapolate(const int x, const int y, const int z) const;
  const Vector3d&cut(const int x, const int y, const int z) const;
  void	convect(SpeedGrid &sg, const double dt, const int steps = 10);
  void	blur(SpeedGrid &sg);
  void	reset() { for (int i(0); i < _m_size; ++i) _m_grid[i] = Vector3d(0.0);}
  void	glRender() const;
  void	eq(const SpeedGrid &sg);
  void	eqSubMul(const double d, const SpeedGrid &w);
  void	eqSubMul(const SpeedGrid &b, const Filter &a, const SpeedGrid &w);
  void	eqSubMul(const SpeedGrid &b, const MultiMatrix &a, const SpeedGrid &w);
  void	eqSubMul(const SpeedGrid &b, const UniMatrix &a, const SpeedGrid &w);
  void	eqSubMul(const SpeedGrid &b, const IsoFilter &a, const SpeedGrid &w);
  void	eqAddMul(const SpeedGrid &b, const double d, const SpeedGrid &w);
  void	eqAddMul(const double d, const SpeedGrid &w);
  void	eqMul(const Filter &a, const SpeedGrid &w);
  void	eqMul(const MultiMatrix &a, const SpeedGrid &w);
  void	eqMul(const UniMatrix &a, const SpeedGrid &w);
  void	eqMul(const IsoFilter &a, const SpeedGrid &w);
  double	norm2() const;
  double	dot(const SpeedGrid &w) const;
  void		keepInside();
  static void	precomputeNghbd(const int x, const int y, const int z,
				const int size);
  static const int	*getNghbd(const int i) { return _s_nghbd + 6 * i; }
};

// ///////////////////////////////////////////////////////////////////////// //
inline void	DivGrid::eqAddMul(const DivGrid	&b,
				  const double	d,
				  const DivGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] = b._m_grid[i] + w._m_grid[i] * d;
}

// ///////////////////////////////////////////////////////////////////////// //
inline void	DivGrid::eqMul(const IsoFilter	&a,
			       const DivGrid	&w)
{
  const double	s(a._m_s), c(a._m_c);
  const int	*nghbd;

  for (int i(0); i < _m_size; ++i)
    {
      /*
      nghbd = getNghbd(i);
      _m_grid[i] = w._m_grid[i] * c + 
	(w._m_grid[nghbd[0]] + w._m_grid[nghbd[1]] + w._m_grid[nghbd[2]] +
	 w._m_grid[nghbd[3]] + w._m_grid[nghbd[4]] + w._m_grid[nghbd[5]]) * s;
      */
      nghbd = getNghbd(i);
      _m_grid[i] = w._m_grid[i] * c;
      for (int j(0); j < 6; ++j)
	if (nghbd[j] != i)
	  _m_grid[i] += w._m_grid[nghbd[j]] * s;
    }
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	DivGrid::norm2() const
{
  double	ret(0.0);

  for (int i(0); i < _m_size; ++i)
    ret += _m_grid[i] * _m_grid[i];
  return ret;
}

// ///////////////////////////////////////////////////////////////////////// //
inline double	DivGrid::dot(const DivGrid &w) const
{
  double	ret(0.0);

  for (int i(0); i < _m_size; ++i)
    ret += _m_grid[i] * w._m_grid[i];
  return ret;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this = b - A * w, where * is a convolution with the inside
///  (ignore this and b on the outside).
// ///////////////////////////////////////////////////////////////////////// //
inline void	DivGrid::eqSubMul(const DivGrid		&b,
				  const IsoFilter	&a,
				  const DivGrid		&w)
{
  const double	s(a._m_s), c(a._m_c);
  const int	*nghbd;

  for (int i(0); i < _m_size; ++i)
    {
      /*
      nghbd = getNghbd(i);
      _m_grid[i] = b._m_grid[i] - w._m_grid[i] * c - 
	(w._m_grid[nghbd[0]] + w._m_grid[nghbd[1]] + w._m_grid[nghbd[2]] +
	 w._m_grid[nghbd[3]] + w._m_grid[nghbd[4]] + w._m_grid[nghbd[5]]) * s;
      */
      nghbd = getNghbd(i);
      _m_grid[i] = b._m_grid[i] - w._m_grid[i] * c;
      for (int j(0); j < 6; ++j)
	if (nghbd[j] != i)
	  _m_grid[i] -= w._m_grid[nghbd[j]] * s;
    }
}

// ///////////////////////////////////////////////////////////////////////// //
inline void	DivGrid::eqAddMul(const double	d,
				  const DivGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] += w._m_grid[i] * d;
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: this -= d * w
// ///////////////////////////////////////////////////////////////////////// //
inline void	DivGrid::eqSubMul(const double	d,
				  const DivGrid	&w)
{
  for (int i(0); i < _m_size; ++i)
    _m_grid[i] -= d * w._m_grid[i];
}

// ///////////////////////////////////////////////////////////////////////// //
/// DESCRIPTION: Computes 3 coeffs for a C1 continuous overshoots-free 
///  interpolation. The first coefficient is the value of the function
///  itself (i.e. the second out of 4).
// ///////////////////////////////////////////////////////////////////////// //
inline Vector3d	coeff(const Vector3d	t[4],
		      const double	x)
{
  Vector3d	ret;

  for (int i(0); i < 3; ++i)
    {
      double		ds(0.5 * (t[2](i) - t[0](i)));
      double		dt(0.5 * (t[3](i) - t[1](i)));
      const double	d(t[2](i) - t[1](i)), d3(3.0 * d);
      double	a1, a2, a3;

      if (d < 0.0)
	{
	  if (d3 > ds)
	    ds = d3;
	  else if (2.0 * ds > d)
	    ds = 0.0;
	  if (d3 > dt)
	    dt = d3;
	  else if (2.0 * dt > d)
	    dt = 0.0;
	}
      else
	{
	  if (d3 < ds)
	    ds = d3;
	  else if (2.0 * ds < d)
	    ds = 0.0;
	  if (d3 < dt)
	    dt = d3;
	  else if (2.0 * dt < d)
	    dt = 0.0;
	}
      a1 = ds;
      a2 = d3 - 2.0 * ds - dt;
      a3 = ds + dt - 2.0 * d;
      ret(i) = (((a3 * x) + a2) * x + a1) * x + t[1](i);
    }
  return ret;
}

// ///////////////////////////////////////////////////////////////////////// //
#endif//__SFSPEEDGRID_HH__
