// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Nov 11 08:48:06 NZDT 2000
// ************************************************************************* //
#include <vector>
#include <list>
#include "gfxTypes.h"

// ************************************************************************* //
class Surfel;
class Edge;
class Triangle;
class Shape;
class Tool;
class ToolPinch;
class ToolVolume;

// ************************************************************************* //
/// DESCRIPTION:
///  isRealTime	- returns true if the sweep is fast enough.
///  sweep	- for sweeping the whole thing.
///  sweepStep	- for sweeping a few steps.
// ************************************************************************* //
class Sweep
{
protected:
  const Matrix4d	&_m_msrc;
  const Transformation	&_m_transf;
  mutable unsigned	_m_istep;
  unsigned		_m_steps;
  BBox3d		_m_bbox;
public:
  Sweep(const Matrix4d &msrc, const Transformation &transf):
    _m_msrc(msrc), _m_transf(transf), _m_istep(0), _m_steps(0) {}
  virtual void	sweep(Shape &s) const;
  virtual void	sweepStep(Shape &s) const;
  virtual void	sweepStep(Surfel &s) const = 0;
  virtual bool	isRealTime() const { return true; }
  virtual bool	beginSweep(Shape &) { return true; }
  bool		isDone() const { return _m_istep >= _m_steps - 1; }
  virtual void	transform(Matrix4d &m) const { m = _m_transf * m; }
  virtual bool	transformGizmo() const { return true; }
  virtual void	query(Shape &s, std::list<Surfel *> &slist) const;
  virtual void	query(Shape &s, std::list<Edge *> &elist) const;
};

// ************************************************************************* //
/// DESCRIPTION: Normal sweep.
// ************************************************************************* //
class SweepSingle: public Sweep
{
  const Tool		&_m_tool;
  Transformation	*_m_tstep;
  std::vector<BBox3d>	_m_sweptbbox;
  std::vector<Matrix4d>	_m_invmsteps;
  std::vector<Matrix4d>	_m_msteps;
public:
  SweepSingle(const Tool &tool, const Matrix4d &msrc,
	      const Transformation &transf);
  virtual ~SweepSingle();
  void	sweepStep(Surfel &s) const;
  void	sweepStep(Point3d &s) const;
  void	sweep(Shape &s) const;
  static void	glRender(const Tool &tool);
};

// ************************************************************************* //
/// DESCRIPTION: Symmetric sweep (X, Y or Z).
// ************************************************************************* //
class SweepSym: public Sweep
{
  const Tool		&_m_tool;
  Transformation	*_m_tstep, *_m_tstepSym;
  Matrix4d		_m_mlogstep, _m_mlogstepSym;
  std::vector<BBox3d>	_m_sweptbbox;
  std::vector<Matrix4d>	_m_invmsteps, _m_invmstepsSym;
  std::vector<Matrix4d>	_m_msteps, _m_mstepsSym;
  bool			_m_intersect;
  const unsigned	_m_sym;
public:
  SweepSym(const Tool &tool, const Matrix4d &msrc,
	   const Transformation &transf, const unsigned sym);
  virtual ~SweepSym();
  void	sweepFarAway(Shape &s) const;
  void	sweep(Shape &s) const;
  void	sweepStep(Surfel &s) const;
  void	sweepStepSym(Point3d &s) const;
  void	sweepStepUni0(Point3d &s) const;
  void	sweepStepUni1(Point3d &s) const;
  static void	glRender(const Tool &tool, const unsigned sym);
};

// ************************************************************************* //
/// DESCRIPTION: Sweeper for a pinch tool.
// ************************************************************************* //
class SweepPinch: public Sweep
{
  const Tool		&_m_tool;
  const Matrix4d	_m_invmsrc;
  Point3d		_m_vSnap;//Pinch center in world coordinates.
  Point3d		_m_vPos[2];//Surface points in world coordinates.
  Vector3d		_m_zAxis;
  mutable std::list<Surfel *>	_m_sharpSurfels;
  mutable Triangle		*_m_toFuse[2];
  mutable std::list<Edge *>	_m_sharpContour[2];
public:
  SweepPinch(const Tool &tool, const Matrix4d &msrc,
	     const Transformation &transf, const unsigned sym);
  virtual ~SweepPinch() {}
  bool	beginSweep(Shape &s);
  void	sweepStep(Surfel &) const {}//empty
  void	sweepPinch(Point3d &p) const;
  void	sweepPinch(Surfel &s) const;
  void	sweepPinch(Shape &s) const;
  void	sweepPull(Point3d &p) const;
  void	sweepPull(Surfel &s) const;
  void	sweepPull(Shape &s) const;
  void	sweepStep(Shape &s) const;
  void	intersect(Shape &s, Point3d vInt[2]) const;
  bool	isRealTime() const { return false; }
  bool	transformGizmo() const { return false; }
};


