// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sun Dec 21 10:18:20 NZDT 2003
// ************************************************************************* //
#include "swpSurfel.h"
#include "swpShape.h"
#include "swpToolProto.h"
#include "swpSweepProto.h"

// ************************************************************************* //
namespace 
{

static const double	g_tab[] = 
  { 1.0, 1.0, 1.0 / 2.0, 1.0 / 3.0, 1.0 / 4.0, 1.0 / 5.0, 1.0 / 6.0, 
    1.0 / 7.0, 1.0 / 8.0, 1.0 / 9.0, 1.0 / 10.0, 1.0 / 11.0, 1.0 / 12.0,
    1.0 / 13.0, 1.0 / 14.0, 1.0 / 15.0, 1.0 / 16.0, 1.0 / 17.0, 1.0 / 18.0,
    1.0 / 19.0, 1.0 / 20.0, 1.0 / 21.0, 1.0 / 22.0, 1.0 / 23.0, 1.0 / 24.0,
    1.0 / 25.0, 1.0 / 26.0, 1.0 / 27.0, 1.0 / 28.0, 1.0 / 29.0, 1.0 / 30.0,
    1.0 / 31.0, 1.0 / 32.0 };

// ************************************************************************* //
}

// ************************************************************************* //
/// DESCRIPTION: Value constructor.
// ************************************************************************* //
SweepProto::SweepProto(const ToolProto		&tool,
		       const Matrix4d		&msrc,
		       const Transformation	&transf,
		       const unsigned		sym):
  Sweep(msrc, transf), _m_tool(tool), 
  _m_nbSpheres((sym? 2: 1) * tool._m_nbSpheres),
  _m_fields(new double[_m_nbSpheres]), _m_sym(sym), _m_r(tool.getr())
{
  //const double		rad((msrc * Vector3d(1.0, 0.0, 0.0)).norm());
  const double		rad((msrc * Vector3d(1.0, 0.0, 0.0)).norm());

  _m_t = Vector3d(transf.mlog() * Point3d(0.0));

  const Vector3d	z(_m_t.unit());
  Vector3d		x((fabs(z.x) < 0.5)? Vector3d(1.0, 0.0, 0.0):
			  Vector3d(0.0, 1.0, 0.0));
  const Point3d	c(msrc * Point3d(0.0));
  Point3d	ci;
  Vector3d	v;

  x = rad * (x - (x * z) * z).unit();

  const Vector3d	y(z ^ x);
  const double		astep(2.0 * M_PI / _m_tool._m_nbSpheres);
  double		sina, cosa;
  
  const double s = ceil(1.875 * 2.0 * _m_t.norm() / rad);
  _m_steps = (s <= 1)? 1: unsigned(s);

  std::cout << "\rsteps = " << _m_steps << " " << std::flush;

  _m_mlog = new Table<Matrix4d>(_m_steps, _m_nbSpheres);
  _m_sphereCenters = new Table<Point3d>(_m_steps, _m_nbSpheres);
  _m_bbox = tool.getBBox(msrc, msrc.inverse());
  _m_c = new Point3d[_m_steps + 1];
  _m_c[_m_steps] = c + _m_t;

  const double	angle((2.0 * _m_t.norm() / 
		       (_m_steps * rad * tool._m_nbSpheres)));

  for (unsigned istep(0); istep < _m_steps; ++istep)
    {
      ci = c + _m_t * (istep / double(_m_steps));
      _m_c[istep] = ci;
      for (unsigned i(0), n(tool._m_nbSpheres); i < n; ++i)
	{
	  cosa = cos(i * astep);
	  sina = sin(i * astep);
	  v = cosa * y - sina * x;
	  (*_m_sphereCenters)(istep, i) = ci + cosa * x + sina * y;
	  (*_m_mlog)(istep, i) = 
	    Rotation(Quaternion::creator(angle, v),
		     (*_m_sphereCenters)(istep, i)).mlog();
	}
    }
}

// ************************************************************************* //
SweepProto::~SweepProto()
{
  delete _m_mlog;
  delete _m_sphereCenters;
  delete[] _m_fields;
  delete[] _m_c;
}

// ************************************************************************* //
/// DESCRIPTION: The field of the spheres forming the tool tool.
// ************************************************************************* //
double	SweepProto::field(const Point3d	&p,
			   const unsigned	i) const
{
  const double	l2((p - (*_m_sphereCenters)(_m_istep, i)).norm2());

  if (_m_tool._m_k2 <= l2)
    return 0.0;

  const double	l(sqrt(l2)), dk(l * _m_tool._m_invk);

  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
/// DESCRIPTION: Number of steps for the combination of rotations.
// ************************************************************************* //
unsigned	SweepProto::nbSteps() const
{
  const double	k((_m_msrc * Vector3d(2.0, 0.0, 0.0)).norm());

  return _m_transf.nbSteps(_m_bbox, k);
}

// ************************************************************************* //
/// DESCRIPTION: Querry the shape with the appropriate symmetry.
// ************************************************************************* //
void	SweepProto::query(Shape		&s,
			   std::list<Surfel *>	&slist) const
{
  s.query(slist, _m_bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Querry the shape with the appropriate symmetry.
// ************************************************************************* //
void	SweepProto::query(Shape		&s,
			   std::list<Edge *>	&elist) const
{
  s.query(elist, _m_bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Blends the rotations. Uses quick exponential.
// ************************************************************************* //
void	SweepProto::swirlStep(Point3d	&p) const
{
  Matrix4d	m(0.0);

  for (unsigned i(0); i < _m_nbSpheres; ++i)
    if ((_m_fields[i] = field(p, i)) == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (_m_fields[i] != 0.0)
      m += (*_m_mlog)(_m_istep, i) * field(p, i);

  const Vector3d	a(m(2, 1), m(0, 2), m(1, 0));
  const Vector3d	b(m(0, 3), m(1, 3), m(2, 3));
  const double		thetak(a.norm());

  if (fabs(a * b) < 1.0e-16 && thetak > 1.0e-8)
    {
      const Vector3d	vk(a / thetak);
      Vector3d	ck;

      if (fabs(a.y) <= fabs(a.x) && fabs(a.z) <= fabs(a.x))
	ck = Vector3d(1.0, (a.y - b.z) / a.x, (a.z + b.y) / a.x);
      else if (fabs(a.z) <= fabs(a.y))
	ck = Vector3d((a.x + b.z) / a.y, 1.0, (a.z - b.x) / a.y);
      else
	ck = Vector3d((a.x - b.y) / a.z, (a.y + b.x) / a.z, 1.0);
      if (ck.norm2() < 1.0e8)
	{
	  Quaternion	q(cos(0.5 * thetak), vk * sin(0.5 * thetak));
	  
	  p -= ck;
	  q(p);
	  p += ck;
	}
      else
	m = m.exp(), m(p);
    }
  else
    m = m.exp(), m(p);
  //p += b;
}

// ************************************************************************* //
void	SweepProto::sweepStep(Point3d	&p) const
{
  Vector3d	u(p - _m_c[_m_istep]);
  const double	rho(u.norm());
  
  if (rho <= 1.0e-4 + _m_r)
    p += _m_t * (1.0 / _m_steps);
  else
    {
      const double l(pow(-_m_r * _m_r * _m_r + rho * rho * rho, 1.0 / 3.0));
      
      u /= rho;
      p = _m_c[_m_istep] + u * l;
      swirlStep(p);
      {
	Vector3d     u(p - _m_c[_m_istep + 1]);
	const double rho(u.norm());
	const double l(pow(_m_r * _m_r * _m_r + rho * rho * rho, 1.0 / 3.0));

	u /= rho;
	p = _m_c[_m_istep + 1] + u * l;
      }
    }
}

// ************************************************************************* //
void	SweepProto::sweepStep(Surfel &s) const
{
  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
  sweepStep(p);
  sweepStep(px);
  sweepStep(py);
  sweepStep(pz);

  Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

  j.eqCommatrix();
  s._m_v *= j;
  s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepProto::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

