// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sat Jan 17 19:30:36 NZDT 2004
// ************************************************************************* //
#include "swpSweepProto.h"
#include "swpToolProto.h"
#include "swpSurfel.h"
#include "swpShape.h"

// ************************************************************************* //
SweepProto::SweepProto(const ToolProto		&tool,
		       const Matrix4d		&msrc,
		       const Transformation	&transf,
		       const unsigned		): Sweep(msrc, transf),
						   _m_tool(tool)
{
  const Vector3d	t(transf.mlog() * Point3d(0.0));

  _m_c = msrc * Point3d(0.0);
  _m_t = t;
  _m_z = t.unit();
  _m_steps = 1;
  _m_tstep = transf.copy();
  _m_angle = RAD2DEG(_m_t.norm());
  *_m_tstep ^= 1.0 / _m_steps;
  _m_bbox = tool.getBBox(msrc, msrc.inverse());
}

// ************************************************************************* //
void	SweepProto::sweepStep(Point3d &p) const
{
  const double	t(tau(p));

  if (t == 1.0)
    (*_m_tstep)(p, t);
  else
    {
      /*
      (*_m_tstep)(p, t);*/
      const double	r(rho(p)), a(alpha(p));
      const Matrix4d	tlog(_m_tstep->mlog());
      const Rotation	rot(Quaternion::creator(_m_angle, (_m_t ^ (p - _m_c))),
			    _m_c);
      const Matrix4d	rlog(rot.mlog());
      const Matrix4d	m((tlog * t + rlog * (r * a)).exp());

      m(p);
    }
}

// ************************************************************************* //
void	SweepProto::sweepStep(Surfel &s) const
{
  const double	t(tau(s._m_p));

  //std::cout << "?" << std::endl;
  if (t == 0.0)
    return;
  
  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  if (t == 1.0)
    (*_m_tstep)(p, t);
  else
    {
      /*
      (*_m_tstep)(p, t);*/
      const double	r(rho(p)), a(alpha(p));
      const Matrix4d	tlog(_m_tstep->mlog());
      const Rotation	rot(Quaternion::creator(_m_angle, (_m_t ^ (p - _m_c))),
			    _m_c);
      const Matrix4d	rlog(rot.mlog());
      const Matrix4d	m((tlog * t + rlog * (r * a)).exp());

      m(p);
    }

      // NORMAL CALCULATION ///////////////////////////////////////////////////
      px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
      sweepStep(px);
      sweepStep(py);
      sweepStep(pz);

      Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

      s._m_detJac = j.det();
      j.eqCommatrix();
      s._m_v *= j;
      s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepProto::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

// ************************************************************************* //
double	SweepProto::tau(const Point3d	&p) const
{
  const double	r2((_m_msrc * Vector3d(1.0, 0.0, 0.0)).norm2());
  const double	l2((p - _m_c).norm2());
  double	dk, l, r;

  if (l2 <= r2)
    return 1.0;
  l = sqrt(l2);
  r = sqrt(r2);
  if (_m_tool._m_k2 <= sq(l - r))
    return 0.0;
  dk = (l - r) * _m_tool._m_invk;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
// TODO: multiply by theta.
// ************************************************************************* //
double	SweepProto::rho(const Point3d	&p) const
{
  const double	r2((_m_msrc * Vector3d(1.0, 0.0, 0.0)).norm2());
  const double	l2((p - _m_c).norm2());
  double	dk, l, r;

  if (l2 <= r2)
    return 0.0;
  r = sqrt(r2);
  if (sq(r + _m_tool._m_k) <= l2)
    return 0.0;
  l = sqrt(l2);
  if (l <= r + 0.5 * _m_tool._m_k)
    {
      dk = ((r + 0.5 * _m_tool._m_k) - l) * _m_tool._m_invk * 2.0;
      return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
    }
  dk = (l - (r + 0.5 * _m_tool._m_k)) * _m_tool._m_invk * 2.0;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
double	SweepProto::alpha(const Point3d	&p) const
{
  const double	a(acos((p - _m_c).unit() * _m_z));
  double	dk;

  if (a < M_PI * 0.5)
    {   
      dk = (M_PI * 0.5 - a) / (M_PI * 0.5);
      return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
    }
  dk = (a - M_PI * 0.5) / (M_PI * 0.5);
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}
