// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sat Jan 17 19:30:36 NZDT 2004
// ************************************************************************* //
#include "swpSweepStokes.h"
#include "swpToolStokes.h"
#include "swpSurfel.h"
#include "swpShape.h"

// ************************************************************************* //
#ifndef SQRT3
#define SQRT3 1.7320508075688772935274
#endif//SQRT3

// ************************************************************************* //
Solver	*SweepStokes::_s_solver(0);

// ************************************************************************* //
SweepStokes::SweepStokes(const ToolStokes	&tool,
			 const Matrix4d		&msrc,
			 const Transformation	&transf,
			 const unsigned		): 
  Sweep(msrc, transf), _m_tool(tool), _m_invmsrc(msrc.inverse())
{
  const Matrix4d	mlog(transf.mlog());
  const int		length(72);
  const double		rad(12.5);
  Vector3d		t, v;
  const double		toolSpeed(2.0);
  double	cosa;

  _m_c = _m_msrc * Point3d(0.0);
  _m_z = Vector3d(mlog * Point3d(0.0)).unit();
 
  const Vector3d real((_m_msrc * Vector3d(1.0, 0.0, 0.0)).unit());
  const Rotation r(Quaternion::creator(acos(_m_z * real), _m_z ^ real), _m_c);

  cosa = _m_z * real;
  if (fabs(cosa - 1.0) < 1.0e-6)
    _m_world2solver = 
      Translation(Vector3d((length - 1.0) * 0.5)).asMatrix() * 
      Scale(rad, Point3d(0.0)).asMatrix() * _m_invmsrc;
  else if (fabs(cosa + 1.0) < 1.0e-6)
    {
      Vector3d	axis((fabs(_m_z.x) < 0.5)? Vector3d(1.0, 0.0, 0.0):
		     Vector3d(0.0, 1.0, 0.0));
      
      axis = rad * (axis - (axis * _m_z) * _m_z).unit();

      Rotation r(Quaternion::creator(M_PI, axis), _m_c);

      _m_world2solver = 
	Translation(Vector3d((length - 1.0) * 0.5)).asMatrix() * 
	Scale(rad, Point3d(0.0)).asMatrix() * _m_invmsrc * r.asMatrix();
    }
  else
    _m_world2solver = 
      Translation(Vector3d((length - 1.0) * 0.5)).asMatrix() * 
      Scale(rad, Point3d(0.0)).asMatrix() * _m_invmsrc * r.asMatrix();

  _m_solver2world = _m_world2solver.inverse();

  if (!_s_solver)
    {
      //_s_solver = new Solver(length, length, length, rad, 0.0, 24.0e-6);
      _s_solver = new Solver(length, length, length, rad, 2.0, 1.0e-6);

      const Matrix4d	w(_m_world2solver * mlog * _m_solver2world);
      const double	k(2.0);//thickness of coating.

      _s_solver->reset();
      for (int x(0); x < length; ++x)
	for (int y(0); y < length; ++y)
	  for (int z(0); z < length; ++z)
	    {
	      const double locr((Point3d(x, y, z) - 
				 Point3d((length - 1.0) * 0.5)).norm() - rad);
	      double dk;

	      if (locr <= 0.0)
		dk = 1.0;
	      else if (k <= locr)
		dk = 0.0;
	      else
		{
		  dk = locr / k;
		  dk = (1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0));
		}
	      (*_s_solver)(x, y, z) = Vector3d(dk * toolSpeed, 0.0, 0.0);
	    }
      _s_solver->solve();
    }
  v = _s_solver->speed(Point3d((length - 1.0) * 0.5));
  t = (_m_world2solver * Vector3d(mlog * Point3d(0.0)));// / toolSpeed;
  _m_steps = max(1, int(ceil(t.norm() / v.norm())));
  _m_len = t.norm() / v.norm() / _m_steps;
  
  std::cout << _m_steps << std::endl;

  _m_tstep = transf.copy();
  (*_m_tstep) ^= 1.0 / _m_steps;
  _m_bbox = tool.getBBox(msrc, _m_invmsrc);
}

// ************************************************************************* //
SweepStokes::~SweepStokes()
{
  delete _m_tstep;
}

// ************************************************************************* //
void	SweepStokes::sweepStep(Point3d &p) const
{
  (*_m_tstep)(p, -double(_m_istep));
  p *= _m_world2solver;
  (*_s_solver)(p, _m_len, 2);
  p *= _m_solver2world;
  (*_m_tstep)(p, double(_m_istep));
}

// ************************************************************************* //
void	SweepStokes::sweepStep(Surfel &s) const
 {
  Point3d	q(s._m_p);
  Vector3d	v;

  (*_m_tstep)(q, -double(_m_istep));
  q *= _m_world2solver;
  v = _s_solver->speed(q);
  if (fabs(v.x + v.y + v.z) < 1.0e-12)
    return;

  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  p = q;
  (*_s_solver)(p, _m_len, 2);
  p *= _m_solver2world;
  (*_m_tstep)(p, double(_m_istep));

  px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;

  sweepStep(px);
  sweepStep(py);
  sweepStep(pz);

  Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

  j.eqCommatrix();
  s._m_v *= j;
  s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepStokes::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

