// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Sun Dec 21 10:18:20 NZDT 2003
// ************************************************************************* //
#include "swpSurfel.h"
#include "swpShape.h"
#include "swpToolVolume.h"
#include "swpSweepVolume.h"
#include "rigidLog.h"

// ************************************************************************* //
namespace 
{

static const double	g_tab[] = 
  { 1.0, 1.0, 1.0 / 2.0, 1.0 / 3.0, 1.0 / 4.0, 1.0 / 5.0, 1.0 / 6.0, 
    1.0 / 7.0, 1.0 / 8.0, 1.0 / 9.0, 1.0 / 10.0, 1.0 / 11.0, 1.0 / 12.0,
    1.0 / 13.0, 1.0 / 14.0, 1.0 / 15.0, 1.0 / 16.0, 1.0 / 17.0, 1.0 / 18.0,
    1.0 / 19.0, 1.0 / 20.0, 1.0 / 21.0, 1.0 / 22.0, 1.0 / 23.0, 1.0 / 24.0,
    1.0 / 25.0, 1.0 / 26.0, 1.0 / 27.0, 1.0 / 28.0, 1.0 / 29.0, 1.0 / 30.0,
    1.0 / 31.0, 1.0 / 32.0 };

// ************************************************************************* //
}

// ************************************************************************* //
/// DESCRIPTION: Value constructor.
// ************************************************************************* //
SweepVolume::SweepVolume(const ToolVolume	&tool,
			 const Matrix4d		&msrc,
			 const Transformation	&transf,
			 const unsigned		sym):
  Sweep(msrc, transf), _m_tool(tool), 
  _m_nbSpheres((sym? 2: 1) * tool._m_nbSpheres),
  _m_fields(new double[_m_nbSpheres]),
  _m_sym(sym)
{
  const double		rad((msrc * Vector3d(1.0, 0.0, 0.0)).norm());
  const Point3d		c(msrc * Point3d(0.0));
  const Vector3d	t(transf.mlog() * Point3d(0.0));
  const Vector3d	z(t.unit());
  Vector3d		x((fabs(z.x) < 0.5)? Vector3d(1.0, 0.0, 0.0):
			  Vector3d(0.0, 1.0, 0.0));
  Point3d	ci;
  Vector3d	v;

  x = rad * (x - (x * z) * z).unit();

  const Vector3d	y(z ^ x);
  const double		astep(2.0 * M_PI / _m_tool._m_nbSpheres);
  double		sina, cosa;
  
  //TODO: translate the tool with the steps!
  //_m_steps = nbSteps();
  //const double	s(ceil(1.875 / k * _m_v.norm()));
  const double s = ceil(1.875 * 2.0 * t.norm() / rad);
  _m_steps = (s <= 1)? 1: unsigned(s);

  std::cout << "\rsteps = " << _m_steps << " " << std::flush;
  //_m_steps = 10;//DEBUG

  _m_mlog = new Table<RigidLog>(_m_steps, _m_nbSpheres);
  //_m_n = new Table<Vector3d>(_m_steps, _m_nbSpheres);
  //_m_m = new Table<Vector3d>(_m_steps, _m_nbSpheres);
  _m_sphereCenters = new Table<Point3d>(_m_steps, _m_nbSpheres);
  _m_bbox = tool.getBBox(msrc, msrc.inverse());
  _m_bbox.fill(c + t * (1.0 + rad));//Added.
  const double	angle((2.0 * t.norm() / 
		       (_m_steps * rad * tool._m_nbSpheres)));
  //const double	angle(2.0 * 45.0 / tool._m_nbSpheres / _m_steps);

  for (unsigned istep(0); istep < _m_steps; ++istep)
    {
      ci = c + t * (istep / double(_m_steps));
      for (unsigned i(0), n(tool._m_nbSpheres); i < n; ++i)
	{
	  Vector3d	nn, cxnn;

	  cosa = cos(i * astep);
	  sina = sin(i * astep);
	  v = (cosa * y - sina * x).unit();
	  (*_m_sphereCenters)(istep, i) = ci + cosa * x + sina * y;
	  nn = angle * v;
	  cxnn = Vector3d((*_m_sphereCenters)(istep, i)) ^ nn;
	  (*_m_mlog)(istep, i) = RigidLog(nn, cxnn);

	  if (sym)
	    {
	      (*_m_sphereCenters)(istep, i + n) =
		(*_m_sphereCenters)(istep, i).sym(sym);
	      nn = -angle * v.sym(sym);
	      cxnn = Vector3d((*_m_sphereCenters)(istep, i + n)) ^ nn;
	      (*_m_mlog)(istep, i + n) = RigidLog(nn, cxnn);
	    }
	}
    }
}

// ************************************************************************* //
SweepVolume::~SweepVolume()
{
  delete _m_mlog;
  delete _m_sphereCenters;
  delete[] _m_fields;
}

// ************************************************************************* //
/// DESCRIPTION: The field of the spheres forming the tool tool.
// ************************************************************************* //
double	SweepVolume::field(const Point3d	&p,
			   const unsigned	i) const
{
  const double	l2((p - (*_m_sphereCenters)(_m_istep, i)).norm2());

  if (_m_tool._m_k2 <= l2)
    return 0.0;

  const double	l(sqrt(l2)), dk(l * _m_tool._m_invk);

  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
/// DESCRIPTION: Number of steps for the combination of rotations.
// ************************************************************************* //
unsigned	SweepVolume::nbSteps() const
{
  const double	k((_m_msrc * Vector3d(2.0, 0.0, 0.0)).norm());

  return _m_transf.nbSteps(_m_bbox, k);
}

// ************************************************************************* //
/// DESCRIPTION: Querry the shape with the appropriate symmetry.
// ************************************************************************* //
void	SweepVolume::query(Shape		&s,
			   std::list<Surfel *>	&slist) const
{
  if (_m_sym)
    {
      BBox3d	qbbox(_m_bbox.sym(_m_sym));
      
      if (_m_bbox.intersects(qbbox))
	{
	  qbbox.fill(_m_bbox);
	  s.query(slist, qbbox);
	}
      else
	s.query(slist, _m_bbox), s.query(slist, qbbox);
    }
  else
    s.query(slist, _m_bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Querry the shape with the appropriate symmetry.
// ************************************************************************* //
void	SweepVolume::query(Shape		&s,
			   std::list<Edge *>	&elist) const
{
  if (_m_sym)
    {
      BBox3d	qbbox(_m_bbox.sym(_m_sym));

      if (_m_bbox.intersects(qbbox))
	{
	  qbbox.fill(_m_bbox);
	  s.query(elist, qbbox);
	}
      else
	s.query(elist, _m_bbox), s.query(elist, qbbox);
    }
  else
    s.query(elist, _m_bbox);
}

// ************************************************************************* //
/// DESCRIPTION: Blends the rotations. Uses quick exponential.
// ************************************************************************* //
void	SweepVolume::sweepStep(Point3d	&p) const
{
  RigidLog	m(0.0);

  for (unsigned i(0); i < _m_nbSpheres; ++i)
    if (_m_fields[i] == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (_m_fields[i] != 0.0)
      m += (*_m_mlog)(_m_istep, i) * field(p, i);
  m.exp(p);
}

// ************************************************************************* //
void	SweepVolume::sweepStep(Surfel &s) const
{
  RigidLog	m(0.0);
  double	sumf(0.0), f;
  
  for (unsigned i(0); i < _m_nbSpheres; ++i)
    if ((f = field(s._m_p, i)))
      {
	_m_fields[i] = f;
	sumf += f;
      }
    else
      _m_fields[i] = 0.0;
  if (sumf == 0.0)
    return;
  for (unsigned i(0); i < _m_nbSpheres; ++i)
    if ((f = _m_fields[i]) == 1.0)
      m += (*_m_mlog)(_m_istep, i);
    else if (f != 0.0)
      m += (*_m_mlog)(_m_istep, i) * f;

  Point3d	&p(s._m_p), px(p), py(p), pz(p);

  m.exp(p);

  px.x += 1.0e-12, py.y += 1.0e-12, pz.z += 1.0e-12;
  sweepStep(px);
  sweepStep(py);
  sweepStep(pz);

  Matrix3d	j((px - p) * 1.0e12, (py - p) * 1.0e12, (pz - p) * 1.0e12);

  j.eqCommatrix();
  s._m_v *= j;
  s._m_v.eqUnit();
}

// ************************************************************************* //
void	SweepVolume::sweep(Shape &s) const
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());

  for (sit = s._m_surfels.begin(); sit != send; ++sit)
    (*sit)->saveSurfel();
  Sweep::sweep(s);
}

