// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Thu Nov 20 17:49:55 NZDT 2003
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include <list>
#include "swpTool.h"
#include "swpSweep.h"
#include "swpShape.h"
#include "swpSurfel.h"

// ************************************************************************* //
bool	Tool::_s_isEmpty(false);

// ************************************************************************* //
Tool::Tool(const double k): _m_m(Matrix4d::identity())
{
  setCoating(k);
}

// ************************************************************************* //
ToolBall::ToolBall(const double	k): Tool(k)
{
}

// ************************************************************************* //
void	ToolBall::glRender() const
{
  glPushMatrix();
  _m_m.glMultMatrix();

  GLUquadric *quad = gluNewQuadric();

  gluSphere(quad, 1.0, 32, 24);
  gluDeleteQuadric(quad);
  glPopMatrix();
}

/* ************************************************************************* **
** DESCRIPTION: Square distance to ellipse. Adapted from Mgc D. Eberly.	     **
**            ^								     **
**            |y							     **
**        _--'b`--_							     **
**       /    |    \							     **
**  ----(-----+-----)a--> x						     **
**       \   c|    /							     **
**        `-._|_.-'							     **
**	      |								     **
** ************************************************************************* */
double	ToolBall::field2Ellipse(const double	a2,
				const double	b2,
				const double	c2,
				const Point3d	&p) const
{
  const double	fA2(a2), fB2(b2), fC2(c2);
  const double	fU2(p(0) * p(0)), fV2(p(1) * p(1)), fW2(p(2) * p(2));
  const double	fA2U2(fA2 * fU2), fB2V2(fB2 * fV2), fC2W2(fC2 * fW2);
  const double	fURatio2(sq(p(0)) / a2);
  const double	fVRatio2(sq(p(1)) / b2);
  const double	fWRatio2(sq(p(2)) / c2);
  const unsigned	iMaxIteration(64);
  const double	fEpsilon(1e-08);
  double	fS, fT, fP, fQ, fR, fP2, fQ2, fR2, fPQ, fPR, fQR, fPQR, fDS;
  double	dk, l2;

  if (fURatio2 + fVRatio2 + fWRatio2 <= 1.0)
    if (_s_isEmpty)
      fT = 0.0;
    else
      return 1.0;
  else
    fT = sqrt(((b2 < a2)? ((c2 < a2)? a2: c2):
	       ((c2 < b2)? b2: c2))) * Vector3d(p).norm();
  for (unsigned i(0); i < iMaxIteration; ++i)
    {
      fP = fT + fA2;
      fQ = fT + fB2;
      fR = fT + fC2;
      fP2 = fP * fP;
      fQ2 = fQ * fQ;
      fR2 = fR * fR;
      fS = fP2 * fQ2 * fR2 - fA2U2 * fQ2 * fR2 
	- fB2V2 * fP2 * fR2 - fC2W2 * fP2 * fQ2;
      if (fabs(fS) < fEpsilon)
	break;
      fPQ = fP * fQ;
      fPR = fP * fR;
      fQR = fQ * fR;
      fPQR = fP * fQ * fR;
      fDS = 2.0 * (fPQR * (fQR + fPR + fPQ) - fA2U2 * fQR * (fQ + fR) -
		   fB2V2 * fPR * (fP + fR) - fC2W2 * fPQ * (fP + fQ));
      fT -= fS / fDS;
    }
  l2 = (p - Point3d(fA2 * p(0) / fP, fB2 * p(1) / fQ,
		    fC2 * p(2) / fR)).norm2();
  if (_m_k2 <= l2)
    return 0.0;
  if (l2 <= 0.0)
    return 1.0;
  dk = sqrt(l2) * _m_invk;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
double	ToolBall::field2Ball(const double	r2,
			     const Point3d	&p) const
{
  const double	l2 = Vector3d(p).norm2();
  double	dk, l;

  if (l2 <= 1.0)
    if (_s_isEmpty)
      {
	l = sqrt(l2);
	if (_m_k2 <= r2 && _m_k2 <= sq(1.0 - l) * r2)
	  return 0.0;
	dk = (1.0 - l) * sqrt(r2) * _m_invk;
	return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
      }
    else
      return 1.0;
  l = sqrt(l2);
  if (_m_k2 <= sq(l - 1.0) * r2)
    return 0.0;
  dk = (l - 1.0) * sqrt(r2) * _m_invk;
  return 1.0 + dk * dk * dk * (dk * (15.0 - 6.0 * dk) - 10.0);
}

// ************************************************************************* //
/// DESCRIPTION: Scalar field.
///  Returns 1 inside, 0 outside, and smoothly interpolates inbetween.
// ************************************************************************* //
double	ToolBall::field(const Point3d	&p,
			const Matrix4d	&m,
			const Matrix4d	&invm) const
{
  const double	a2(m.getx().norm2()), b2(m.gety().norm2());
  const double	c2(m.getz().norm2());

  if (fabs(a2 - b2) < 1.0e-6 && fabs(c2 - b2) < 1.0e-6)
    return field2Ball(c2, invm * p);

  const Matrix4d mm(m.getx().unit(), m.gety().unit(),
		    m.getz().unit(), m.geto());

  return field2Ellipse(a2, b2, c2, mm.inverse() * p);
}

// ************************************************************************* //
/// DESCRIPTION: Returns the tool's bbox in world coordinates.
// ************************************************************************* //
BBox3d	ToolBall::getBBox(const Matrix4d &m,
			  const Matrix4d &) const
{
  const Matrix4d mm(m.getx().unit(), m.gety().unit(),
		    m.getz().unit(), m.geto());

  const double	a(mm.getx().norm());
  const double	b(mm.gety().norm());
  const double	c(mm.getz().norm());
  
  BBox3d	ret(Point3d(0.0));

  ret.fill(Point3d(a, b, c));
  ret.fill(Point3d(-a, -b, -c));
  ret = m * ret;
  ret.expand(_m_k);
  return ret;
}

// ************************************************************************* //
/// DESCRIPTION: Returns a sweep for that tool.
// ************************************************************************* //
Sweep	*ToolBall::newSweep(const int			sym,
			    const Matrix4d		&msrc,
			    const Transformation	&transf) const
{
  if (sym == 0)
    return new SweepSingle(*this, msrc, transf);
  return new SweepSym(*this, msrc, transf, sym);
}

// ************************************************************************* //
/// DESCRIPTION: Colorizes the shape with the field of the tool.
// ************************************************************************* //
void	Tool::colorize(const Matrix4d	&msrc,
		       Shape		&s,
		       const unsigned	sym)
{
  std::list<Surfel *>::iterator	sit, send(s._m_surfels.end());
  const Matrix4d	invmsrc = msrc.inverse();

  if (sym == 0)
    for (sit = s._m_surfels.begin(); sit != send; ++sit)
      (*sit)->_m_color = field((*sit)->_m_p, msrc, invmsrc);
  else
    {
      const Matrix4d	&m(Matrix4d::sym(sym));
      
      for (sit = s._m_surfels.begin(); sit != send; ++sit)
	(*sit)->_m_color = max(field((*sit)->_m_p, msrc, invmsrc),
			       field(m * (*sit)->_m_p, msrc, invmsrc));
    }
}
