// ************************************************************************* //
/// AUTHOR: Alexis Angelidis
/// STARTED: Tue Nov 25 13:45:28 NZDT 2003
// ************************************************************************* //
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/glu.h>
#include "swpToolPinch.h"
#include "swpSweep.h"

// ************************************************************************* //
/// DESCRIPTION:|
///         ___|___rOut
///      _-'   |   '-_
///    ,'     _|_     ',aIn
///   / `  _-'rIn '-_  / `
///   `,_``    |    '/_ ,'aOut
///    `` `    |    /  /
///     `. `   |   / ,'
///       ` `  |  / /
///        ',` | /,'
///          ``|//
///           `.' dist (animated)
///            |
// ************************************************************************* //
ToolPinch::ToolPinch(const double k): 
  Tool(k), _m_rIn(0.1), _m_rOut(1.0),
  _m_aIn(0.1 * M_PI), _m_aOut(M_PI), _m_dist(1.0)
{
}

// ************************************************************************* //
BBox3d	ToolPinch::getBBox(const Matrix4d &m,
			   const Matrix4d &) const
{
  BBox3d	bbox(Point3d(-1.0, -1.0, -(1.0 + _m_dist)),
		     Point3d(1.0, 1.0, 1.0 + _m_dist));

  return m * bbox;
}

// ************************************************************************* //
double	ToolPinch::field(const Point3d	&p,
			 const Matrix4d	&m,
			 const Matrix4d	&invm) const
{
  return 0.5 * (fabs(fieldPinch(p, m, invm, 0)) + fieldPull(p, m, invm, 0));
}

// ************************************************************************* //
double	ToolPinch::fieldPinch(const Point3d	&n,
			      const Matrix4d	&,
			      const Matrix4d	&invm,
			      const double	u) const
{
  const double	aIn((1.0 - u) * _m_aIn);
  const double	aOut((1.0 - u) * _m_aOut);
  const Point3d	p(invm * n);
  const double	dist((1.0 - u) * _m_dist);
  const bool	isNegative(p(2) < 0.0);
  const Point3d	q(p(0), p(1), isNegative? (-p(2) - dist): (p(2) - dist));

  if (q(2) <= 0.0)
    return 0.0;

  const double	r2(Vector3d(q).norm2());
  
  if (_m_rOut * _m_rOut <= r2)
    return 0.0;

  const double	r(sqrt(r2)), a(acos(q(2) / r));

  if (aOut <= a)
    return 0.0;
  if (r <= _m_rIn && a <= aIn)
    return isNegative? -1.0: 1.0;

  double	rho, alpha;

  if (r2 <= _m_rIn * _m_rIn)
    rho = 1.0;
  else
    {
      rho = (r - _m_rOut) / (_m_rIn - _m_rOut);
      rho = rho * rho * (3.0 - 2.0 * rho);
    }
  if (a <= aIn)
    alpha = 1.0;
  else 
    {
      alpha = (a - aOut) / (aIn - aOut);
      alpha = alpha * alpha * (3.0 - 2.0 * alpha);
    }
  return isNegative? -(alpha * rho): (alpha * rho);
}

// ************************************************************************* //
double	ToolPinch::fieldPull(const Point3d	&q,
			     const Matrix4d	&,
			     const Matrix4d	&invm,
			     const double	) const
{
  const Point3d	p(invm * q);
  const double	coat(0.5);
  double	d2, dDivCoat;

  if (p(2) < -1.0)
    d2 = sq(p(0)) + sq(p(1)) + sq(p(2) + 1.0);
  else if (1.0 < p(2))
    d2 = sq(p(0)) + sq(p(1)) + sq(p(2) - 1.0);
  else
    d2 = sq(p(0)) + sq(p(1));
  if (coat * coat <= d2)
    return 0.0;
  dDivCoat = sqrt(d2) / coat;
  return 1.0 - dDivCoat * dDivCoat * (3.0 - 2.0 * dDivCoat);
}

// ************************************************************************* //
double	ToolPinch::field(const Point3d	&p,
			 const Matrix4d	&m,
			 const Matrix4d	&invm,
			 const unsigned	istep) const
{
  static const double	steps2 = 512;

  if (istep < steps2)
    return fieldPinch(p, m, invm, istep / double(steps2));
  return fieldPull(p, m, invm, (istep - steps2) / double(steps2));
}

// ************************************************************************* //
void	ToolPinch::glRender() const
{
  static bool	isListInit(false);
  static GLuint	list;
  GLUquadricObj	*qobj;

  if (!isListInit)
    {
      isListInit = true;
      list = glGenLists(1);
      glNewList(list, GL_COMPILE);

      glPushAttrib(GL_TRANSFORM_BIT);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glTranslated(0.0, 0.0, -2.0);

      qobj = gluNewQuadric();
      gluQuadricDrawStyle(qobj, GLU_FILL);
      gluQuadricNormals(qobj, GLU_SMOOTH);
      glTranslated(0.0, 0.0, 0.8);
      glScaled(0.2, 0.2, 0.2);
      gluCylinder(qobj, 0.33, 0.0, 1.0, 16, 1);
      gluDisk(qobj, 0.33, 0.0, 16, 1);
      gluDeleteQuadric(qobj);
      glPopMatrix();//GL_MODELVIEW
      glPopAttrib();//GL_TRANSFORM_BIT
      glEndList();
    }
  {
    glPushMatrix();
    _m_m.glMultMatrix();
    glCallList(list);
    glRotated(180.0, 0.0, 1.0, 0.0);
    glCallList(list);
    glPopMatrix();//GL_MODELVIEW
  }
}

// ************************************************************************* //
Sweep	*ToolPinch::newSweep(const int			sym,
			     const Matrix4d		&msrc,
			     const Transformation	&transf) const
{
  return new SweepPinch(*this, msrc, transf, sym);
}

